#!/usr/bin/env python3

"""
Migration script for hangoutsbot -> IMMP.

Transfers and converts data for:
- nicknames (identities)
- syncrooms (syncs)
- SlackRTM* (syncs, identities)
- telesync* (syncs, identities)
- tldr (notes)

* API keys in config.json will be used during migration to discover the bot remote user's IDs.
"""

from argparse import ArgumentParser, FileType
from collections import defaultdict
from functools import partial
import json
import logging
import os.path
import re

import anyconfig
from playhouse.db_url import connect
from requests import Session

from immp import Nullable, Optional, Schema
from immp.hook.alerts import SubTrigger
from immp.hook.database import BaseModel
from immp.hook.identitylocal import IdentityGroup, IdentityLink
from immp.hook.notes import Note


log = logging.getLogger(__name__)


class _Schema:

    config = Schema({
        "plugins": [str],
        Optional("sync_rooms", list): [[str]],
        Optional("slackrtm"): Nullable({  # rewrite
            "syncs": [{
                "channel": [str, str],
                "hangout": str
            }],
            "teams": {str: {
                "token": str,
                "admins": [str]
            }}
        }),
        Optional("telesync"): Nullable({
            "api_key": str
        }),
        Optional("forwarding", default=dict): {str: {
            "targets": [str]
        }}
    })

    user = Schema({
        Optional("_hangups", lambda: {"is_self": False}): {"is_self": bool},
        Optional("nickname", ""): str,
        Optional("keywords", list): [str]
    })

    memory = Schema({
        "convmem": {str: {
            "title": str
        }},
        "user_data": {str: user},
        Optional("slackrtm", dict): {str: {
            "identities": {
                "hangouts": {str: str},
                "slack": {str: str}
            }
        }},
        Optional("profilesync", lambda: {"ho2tg": {}}): {  # telesync
            "ho2tg": {str: str}  # HO: TG or HO: "VERIFY000..."
        },
        Optional("telesync", lambda: {"ho2tg": {}}): {
            "ho2tg": {str: str}  # HO: TG
        },
        Optional("tldr", dict): {  # HO: timestamp: text
            str: {str: str}
        }
    })


class RevDict(dict):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.inverse = {}
        for key, value in self.items():
            if value in self.inverse:
                raise KeyError(value)
            self.inverse[value] = key

    def __setitem__(self, key, value):
        if value in self.inverse and self.inverse[value] != key:
            raise KeyError(value)
        if key in self:
            del self.inverse[self[key]]
        super().__setitem__(key, value)
        self.inverse[value] = key

    def __delitem__(self, key):
        del self.inverse[self[key]]
        super().__delitem__(key)


class MultiRevDict(dict):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.inverse = defaultdict(list)
        for key, value in self.items():
            self.inverse[value].append(key)

    def __setitem__(self, key, value):
        if key in self:
            self.inverse[self[key]].remove(key)
        super().__setitem__(key, value)
        self.inverse[value].append(key)

    def __delitem__(self, key):
        value = self[key]
        self.inverse[value].remove(key)
        if value in self.inverse and not self.inverse[value]:
            del self.inverse[value]
        super().__delitem__(key)


class Data:

    def __init__(self, config, memory, db_url, path):
        self.config = config
        self.memory = memory
        self.database = connect(db_url)
        BaseModel._meta.database.initialize(self.database)
        # Find the bot self user to construct the network ID, needed by other hooks.
        for uid, user in memory["user_data"].items():
            if user["_hangups"]["is_self"]:
                log.debug("Bot user ID: {}".format(uid))
                self.network_id = "hangouts:{}".format(uid)
                break
        else:
            raise RuntimeError("Can't identify bot user ID")
        # Assorted data structures to help with lookups during the migration.
        self.plugs = {"hangouts": {"path": "immp.plug.hangouts.HangoutsPlug",
                                   "config": {"cookie": os.path.join(path, "cookies.json")}}}
        self.channels = RevDict()
        self.hooks = {"db": {"path": "immp.hook.database.DatabaseHook",
                             "config": {"url": db_url}}}
        self.identities = MultiRevDict()
        self.syncs = MultiRevDict()
        self.forwards = defaultdict(list)
        self.subs = {}
        # Internal counters for unique name generation.
        self.user_count = 0
        self.session = Session()

    # Assorted utility methods used during the migration.

    def get_nickname(self, uid):
        try:
            nick = self.memory["user_data"][uid]["nickname"]
        except KeyError:
            log.warning("User missing from user_data: {}".format(uid))
            self.memory["user_data"][uid] = _Schema.user({})
            nick = None
        if nick:
            log.debug("Got existing nickname: {} -> {}".format(uid, nick))
        else:
            while True:
                self.user_count += 1
                nick = "no-name-{}".format(self.user_count)
                if not any(user.get("nickname") and user["nickname"] == nick
                           for user in self.memory["user_data"].values()):
                    break
            # Apply it to our copy of memory for later lookups of the same user.
            self.memory["user_data"][uid]["nickname"] = nick
            log.debug("Assigned new nickname: {} -> {}".format(uid, nick))
        return nick

    def format_title(self, prefix, title):
        if not title:
            return None
        formatted = re.sub(r"[^a-z0-9]+", "-", title.replace("'", ""), flags=re.I).strip("-")
        return "{}:{}".format(prefix, formatted)

    def hangout_title(self, chat):
        # Prefer a name based on the current conv title.
        return (self.format_title("HO", self.memory["convmem"][chat]["title"])
                if chat in self.memory["convmem"] else None)

    def add_channel(self, plug, source, name_getter=None, name_fallback=None):
        if (plug, source) in self.channels.inverse:
            # Already exists under another name.
            name = self.channels.inverse[(plug, source)]
            log.debug("Preferring existing channel: {} -> {}/{}".format(name, plug, source))
        else:
            name = (name_getter(source) if name_getter else None) or name_fallback
            unique = name
            count = 0
            while name in self.channels:
                count += 1
                name = "{}-{}".format(unique, count)
            self.channels[name] = (plug, source)
            log.debug("Assigned new channel: {} -> {}/{}".format(name, plug, source))
        return name

    def add_sync(self, label, *channels):
        for channel in channels:
            if channel in self.syncs:
                # We've already got a sync for that channel.
                # Merge all of those channels into this sync.
                dupe = self.syncs[channel]
                log.debug("Channel already in sync: {} -> {}".format(channel, dupe))
                # Copy the list as it will be emptied during iteration.
                for synced in list(self.syncs.inverse[dupe]):
                    log.debug("Merging channel into sync: {} -> {}".format(synced, label))
                    self.syncs[synced] = label
                log.debug("Removing now-empty sync: {}".format(dupe))
                del self.syncs.inverse[dupe]
            else:
                log.debug("Adding channel to sync: {} -> {}".format(channel, label))
                self.syncs[channel] = label

    def get_synced(self, plug, source):
        try:
            name = self.channels.inverse[(plug, source)]
        except KeyError:
            return (plug, source)
        # Get the containing sync channel if one exists.
        if name in self.syncs:
            return ("sync", self.syncs[name])
        else:
            return (plug, source)

    def add_forward(self, source, *targets):
        log.debug("Adding channels to forwards: {} -> {}".format(source, ", ".join(targets)))
        self.forwards[source].extend(targets)

    # Migrate Hangouts identities for anyone with a nickname set.

    def ho_identities(self):
        for uid, user in self.memory["user_data"].items():
            if user["nickname"]:
                self.identities[(self.network_id, uid)] = user["nickname"]

    # Migrate syncrooms syncs.

    def syncrooms_syncs(self):
        for i, synced in enumerate(self.config["sync_rooms"]):
            channels = [self.add_channel("hangouts", ho, self.hangout_title,
                                         "HO:syncrooms:{}-{}".format(i, j))
                        for j, ho in enumerate(synced)]
            log.debug("Adding Hangouts sync: {}".format(", ".join(channels)))
            self.add_sync("syncrooms-{}".format(i), *channels)

    # Migrate SlackRTM syncs and identities.

    def slackrtm_api(self, name, endpoint, **kwargs):
        kwargs["token"] = self.config["slackrtm"]["teams"][name]["token"]
        data = self.session.get("https://slack.com/api/{}".format(endpoint), params=kwargs).json()
        if data["ok"]:
            return data
        else:
            raise ValueError(data)

    def slackrtm_network_id(self, name):
        data = self.slackrtm_api(name, "auth.test")
        id = "slack:{}:{}".format(data["team_id"], data["user_id"])
        log.debug("Generated Slack network ID: {}".format(id))
        return id

    def slackrtm_title(self, name, channel):
        try:
            data = self.slackrtm_api(name, "conversations.info", channel=channel)
        except ValueError:
            return None
        else:
            return self.format_title("Slack", data["channel"].get("name"))

    def slackrtm_syncs(self):
        for name, team in self.config["slackrtm"]["teams"].items():
            plug = "slack-{}".format(name)
            self.plugs[plug] = {"path": "immp.plug.slack.SlackPlug",
                                "config": {"token": team["token"]}}
            # IMMP requires channel names, but we don't have much to go on.
            for i, sync in enumerate(self.config["slackrtm"]["syncs"]):
                team, channel = sync["channel"]
                hname = self.add_channel("hangouts", sync["hangout"], self.hangout_title,
                                         "HO:SlackRTM:{}".format(i))
                sname = self.add_channel(plug, channel, partial(self.slackrtm_title, name),
                                         "Slack:SlackRTM:{}".format(i))
                log.debug("Adding Slack sync: {} <-> {}".format(hname, sname))
                self.add_sync("Sync:{}".format(hname.replace("HO:", "")), hname, sname)

    def slackrtm_identities(self):
        for name, data in self.memory["slackrtm"].items():
            network = self.slackrtm_network_id(name)
            for ho, sk in data["identities"]["hangouts"].items():
                if data["identities"]["slack"].get(sk) == ho:
                    nick = self.get_nickname(ho)
                    log.debug("Adding Slack identity: {} -> {}, {}/{}".format(nick, ho, name, sk))
                    self.identities[(self.network_id, ho)] = self.identities[(network, sk)] = nick

    # Migrate telesync syncs and identities.

    def telegram_api(self, endpoint, **kwargs):
        # Calls may be rate-limited and hang for 10-15 seconds after a large number of requests.
        data = self.session.get("https://api.telegram.org/bot{}/{}"
                                .format(self.config["telesync"]["api_key"], endpoint),
                                params=kwargs).json()
        if data["ok"]:
            return data
        else:
            raise ValueError(data)

    def telegram_network_id(self):
        data = self.telegram_api("getMe")
        id = "telegram:{}".format(data["result"]["id"])
        log.debug("Generated Telegram network ID: {}".format(id))
        return id

    def telegram_title(self, chat):
        if int(chat) > 0:
            return None  # Private chat, no title.
        try:
            data = self.telegram_api("getChat", chat_id=chat)
        except ValueError:
            return None
        else:
            return self.format_title("TG", data["result"].get("title"))

    def telesync_syncs(self):
        self.plugs["telegram"] = {"path": "immp.plug.telegram.TelegramPlug",
                                  "config": {"token": self.config["telesync"]["api_key"]}}
        for i, (ho, tg) in enumerate(self.memory["telesync"]["ho2tg"].items()):
            hname = self.add_channel("hangouts", ho, self.hangout_title,
                                     "HO:telesync:{}".format(i))
            tname = self.add_channel("telegram", tg, self.telegram_title,
                                     "TG:telesync:{}".format(i))
            log.debug("Adding Telegram sync: {} <-> {}".format(hname, tname))
            self.add_sync("Sync:{}".format(hname.replace("HO:", "")), hname, tname)

    def telesync_identities(self):
        network = self.telegram_network_id()
        for ho, tg in self.memory["profilesync"]["ho2tg"].items():
            # Ignore incomplete profile syncs.
            if not (ho.startswith("VERIFY") or tg.startswith("VERIFY")):
                nick = self.get_nickname(ho)
                log.debug("Adding Telegram identity: {} -> {}, {}".format(nick, ho, tg))
                self.identities[(self.network_id, ho)] = self.identities[(network, tg)] = nick

    # Migrate Hangouts forwarding.

    def forwarding(self):
        for i, (source, forward) in enumerate(self.config["forwarding"].items()):
            hname = self.add_channel("hangouts", source, self.hangout_title,
                                     "HO:forward:{}-source".format(i))
            hangouts = [self.add_channel("hangouts", target, self.hangout_title,
                                         "HO:forward:{}-{}".format(i, j))
                        for j, target in enumerate(forward["targets"])]
            channels = []
            for ho in hangouts:
                channel = self.syncs.get(ho, ho)
                if channel not in channels:
                    channels.append(channel)
            log.debug("Adding forward: {} -> {} channel(s)".format(hname, len(channels)))
            self.add_forward(hname, *channels)

    # Migrate subscription keywords to Hangouts users.

    def keywords(self):
        tg_network_id = self.telegram_network_id()
        for uid, user in self.memory["user_data"].items():
            if not user["keywords"]:
                continue
            key = (self.network_id, uid)
            try:
                nick = self.identities[key]
            except KeyError:
                identities = [key]
            else:
                identities = [(network, uid) for network, uid in self.identities.inverse[nick]
                              if network in (self.network_id, tg_network_id)]
            for identity in identities:
                self.subs[identity] = filter(None, (re.sub(r"[^\w ]", "", sub).lower()
                                                    for sub in user["keywords"]))

    # Migrate tldrs, assign to synced conversations if relevant.

    def tldr(self):
        self.hooks["notes"] = {"path": "immp.hook.notes.NotesHook"}
        self.database.drop_tables([Note], safe=True)
        self.database.create_tables([Note], safe=True)
        syncs = set()
        for ho, tldr in self.memory["tldr"].items():
            if not tldr:
                continue
            plug, source = self.get_synced("hangouts", ho)
            if plug == "hangouts":
                network = self.network_id
            else:
                # Using a synced channel.
                if source in syncs:
                    # Already processed this tldr via syncrooms sharing.
                    log.debug("Skipping duplicated synced tldr: {}".format(source))
                    continue
                network = "sync:sync"
                syncs.add(source)
            log.debug("Adding {} note(s) to channel: {}/{}".format(len(tldr), plug, source))
            with self.database.atomic():
                for ts, note in sorted(tldr.items()):
                    Note.create(timestamp=int(float(ts)), network=network,
                                channel=source, text=note)

    # Putting it all together now.

    def migrate_all(self):
        self.ho_identities()
        if self.config["sync_rooms"]:
            self.syncrooms_syncs()
        if self.config["slackrtm"]:
            self.slackrtm_syncs()
        if self.memory["slackrtm"]:
            self.slackrtm_identities()
        if self.config["telesync"] and self.memory["telesync"]["ho2tg"]:
            self.telesync_syncs()
        if self.memory["profilesync"]["ho2tg"]:
            self.telesync_identities()
        if self.config["forwarding"]:
            self.forwarding()
        self.keywords()
        if self.memory["tldr"]:
            self.tldr()

    def compile_identities(self):
        self.database.drop_tables([IdentityGroup, IdentityLink], safe=True)
        self.database.create_tables([IdentityGroup, IdentityLink], safe=True)
        self.hooks["identity"] = {"path": "immp.hook.identity.IdentityHook",
                                  "config": {"instance": 1, "plugs": list(self.plugs)}}
        with self.database.atomic():
            for nick, links in self.identities.inverse.items():
                # Invalid password hash by default.
                # Users must `id-password` before they can manage their identities.
                group = IdentityGroup.create(instance=1, name=nick, pwd="")
                for network, user in links:
                    IdentityLink.create(group=group, network=network, user=user)

    def compile_syncs(self):
        identities = "identity" if self.identities else None
        if self.syncs:
            self.hooks["sync"] = {"path": "immp.hook.sync.SyncHook",
                                  "config": {"plug": "sync-migrated",
                                             "channels": dict(self.syncs.inverse),
                                             "identities": identities}}
        if self.forwards:
            self.hooks["forward"] = {"path": "immp.hook.sync.ForwardHook",
                                     "config": {"channels": dict(self.forwards),
                                                "identities": identities}}

    def compile_subs(self):
        self.database.drop_tables([SubTrigger], safe=True)
        self.database.create_tables([SubTrigger], safe=True)
        self.hooks["subs"] = {"path": "immp.hook.alerts.SubscriptionsHook",
                              "config": {"groups": ["migrated"]}}
        with self.database.atomic():
            for (network, user), subs in self.subs.items():
                for text in subs:
                    SubTrigger.create(network=network, user=user, text=text)

    def compile_commands(self):
        commands = {"groups": ["migrated"],
                    "hooks": ["commands"] + list(hook for hook in self.hooks if not hook == "db")}
        self.hooks["commands"] = {"path": "immp.hook.command.CommandHook",
                                  "config": {"prefix": ["/bot "],
                                             "mapping": {"migrated": commands}}}

    def make_config(self):
        if self.identities:
            self.compile_identities()
        if self.syncs or self.forwards:
            self.compile_syncs()
        if self.subs:
            self.compile_subs()
        self.compile_commands()
        return {"plugs": self.plugs,
                "channels": {name: {"plug": plug, "source": source}
                             for name, (plug, source) in self.channels.items()},
                "groups": {"migrated": {"anywhere": list(self.plugs.keys())}},
                "hooks": self.hooks}


def main(args):
    config = _Schema.config(json.load(args.config))
    memory = _Schema.memory(json.load(args.memory))
    data = Data(config, memory, args.database, os.path.dirname(args.config.name))
    data.migrate_all()
    anyconfig.dump(data.make_config(), args.output)


def entrypoint():
    logging.basicConfig(level=logging.DEBUG)
    parser = ArgumentParser(add_help=False)
    parser.add_argument("config", type=FileType("r"))
    parser.add_argument("memory", type=FileType("r"))
    parser.add_argument("output")
    parser.add_argument("database")
    args = parser.parse_args()
    main(args)


if __name__ == "__main__":
    entrypoint()



"""
Interact with and debug a running app in the console.

Synchronous
~~~~~~~~~~~

Config:
    all (bool):
        ``True`` to process any message, ``False`` (default) to restrict to defined channels.
    console (str):
        Use a different embedded console.  By default, :meth:`code.interact` is used, but set this
        to ``ptpython`` (requires the `ptpython <https://github.com/jonathanslenders/ptpython>`_
        Python module) for a more functional shell.

When a new message is received, a console will launch in the terminal where your app is running.
The variables :data:`channel` and :data:`msg` are defined in the local scope, whilst :data:`self`
refers to the shell hook itself.

.. warning::
    The console will block all other running tasks; notably, all plugs will be unable to make any
    progress whilst the console is open.

Asynchronous
~~~~~~~~~~~~

Config:
    port (int):
        Port to bind the console on.  Once running, one can connect using e.g. netcat.  See
        `aioconsole's docs <https://aioconsole.readthedocs.io/en/latest/#serving-the-console>`_
        for more info.
    buffer (int):
        Number of received messages to keep cached for later inspection.  If unset (default), no
        buffer will be available.  Otherwise, when a new message comes in and the queue is full,
        the oldest message will be discarded.  Set to ``0`` for an unlimited buffer, not
        recommended on production deployments.

At startup, a console will be launched on the given port.  You can connect to it from a separate
terminal, for example::

    $ rlwrap nc localhost $PORT

Use of ``rlwrap`` provides you with readline-style keybinds, such as ↑ and ↓ to navigate through
previous commands.  The variables :data:`shell` and :data:`host` are defined, refering to the shell
hook and the running :class:`.Host` respectively.  This hook also maintains a cache of messages as
they're received, accessible via :attr:`.AsyncShellHook.buffer`.

.. note::
    This hook requires the `aioconsole <https://aioconsole.readthedocs.io>`_ Python module.

.. warning::
    The console will be accessible on a locally bound port without authentication.  Do not use on
    shared or untrusted systems, as the host and all connected plugs are exposed.
"""

import code
from collections import deque
from functools import partial
import logging
from pprint import pformat

import immp


try:
    import ptpython.repl
except ImportError:
    ptpython = None

try:
    import aioconsole
except ImportError:
    aioconsole = None


log = logging.getLogger(__name__)


class ShellHook(immp.ResourceHook):
    """
    Hook to start a Python shell when a message is received.
    """

    schema = immp.Schema({immp.Optional("all", False): bool,
                          immp.Optional("console"): immp.Nullable("ptpython")})

    def __init__(self, name, config, host):
        super().__init__(name, config, host)
        if self.config["console"] == "ptpython":
            if ptpython:
                log.debug("Using ptpython console")
                self.console = self._ptpython
            else:
                raise immp.PlugError("'ptpython' module not installed")
        else:
            log.debug("Using native console")
            self.console = self._code

    @staticmethod
    def _ptpython(local):
        ptpython.repl.embed(globals(), local)

    @staticmethod
    def _code(local):
        code.interact(local=dict(globals(), **local))

    async def on_receive(self, sent, source, primary):
        await super().on_receive(sent, source, primary)
        if sent.channel in self.host.channels or self.config["all"]:
            log.debug("Entering console: %r", sent)
            self.console(locals())


class AsyncShellHook(immp.ResourceHook):
    """
    Hook to launch an asynchonous console alongside a :class:`.Host` instance.

    Attributes:
        buffer (collections.deque):
            Queue of recent messages, the length defined by the ``buffer`` config entry.
        last ((.SentMessage, .Message) tuple):
            Most recent message received from a connected plug.
    """

    schema = immp.Schema({"bind": immp.Any(str, int),
                          immp.Optional("buffer"): immp.Nullable(int)})

    def __init__(self, name, config, host):
        super().__init__(name, config, host)
        if not aioconsole:
            raise immp.PlugError("'aioconsole' module not installed")
        self.buffer = None
        self._server = None

    @property
    def last(self):
        return self.buffer[-1] if self.buffer else None

    async def start(self):
        await super().start()
        if self.config["buffer"] is not None:
            self.buffer = deque(maxlen=self.config["buffer"] or None)
        if isinstance(self.config["bind"], str):
            log.debug("Launching console on socket %s", self.config["bind"])
            bind = {"path": self.config["bind"]}
        else:
            log.debug("Launching console on port %d", self.config["bind"])
            bind = {"port": self.config["bind"]}
        self._server = await aioconsole.start_interactive_server(factory=self._factory, **bind)

    async def stop(self):
        await super().stop()
        self.buffer = None
        if self._server:
            log.debug("Stopping console server")
            self._server.close()
            self._server = None

    @staticmethod
    def _pprint(console, obj):
        console.print(pformat(obj))

    def _factory(self, streams=None):
        context = {"host": self.host, "shell": self, "immp": immp}
        console = aioconsole.AsynchronousConsole(locals=context, streams=streams)
        context["pprint"] = partial(self._pprint, console)
        return console

    async def on_receive(self, sent, source, primary):
        await super().on_receive(sent, source, primary)
        if self.buffer is not None:
            self.buffer.append((sent, source))



"""
Connect to `Telegram <https://telegram.org>`_ as a bot.

Config:
    token (str):
        Telegram bot token for the bot API.
    api-id (int):
        Optional Telegram application ID for the MTProto API.
    api-hash (str):
        Corresponding Telegram application secret.
    client-updates (bool):
        ``True`` (when API credentials are configured) to listen for messages over the MTProto
        connection instead of long-polling with the bot API.

        This allows you to maintain multiple connections with a single bot account, however it will
        not mark messages as read within the bot API.  This means if you switch back to the bot API
        at a later date, you'll receive a backlog of any messages not yet picked up.
    session (str):
        Optional path to store a session file, used to cache access hashes.

Bots can be created by talking to `@BotFather <https://t.me/BotFather>`_.  Use the ``/token``
command to retrieve your API token.  You should also call ``/setprivacy`` to grant the bot
permission to see all messages as they come in.

Telegram bots are rather limited by the bot API.  To make use of some features (including lookup of
arbitrary users and listing members in chats), you'll need to combine bot authorisation with an
MTProto application.  This is done via `app credentials <https://my.telegram.org/apps>`_ applied to
a bot session -- the bot gains some extra permissions from the "app", but accesses them itself.

Note that most objects in Telegram cannot be retrieved by ID until you've "seen" them via other
methods.  With a session file, you need only do this once, after which the reference to it (the
"access hash") is cached.

.. note::
    Use of app features requires the `telethon <https://telethon.readthedocs.io/en/latest/>`_
    Python module.
"""

from asyncio import CancelledError, TimeoutError, ensure_future, gather, sleep
from collections import defaultdict
from datetime import datetime, timezone
import logging

from aiohttp import ClientError, ClientResponseError, FormData

import immp


try:
    from telethon import TelegramClient, events, tl
    from telethon.errors import BadRequestError
    from telethon.sessions import SQLiteSession
    from telethon.utils import pack_bot_file_id
except ImportError:
    TelegramClient = SQLiteSession = None


log = logging.getLogger(__name__)


class _Schema:

    config = immp.Schema({"token": str,
                          immp.Optional("api-id"): immp.Nullable(int),
                          immp.Optional("api-hash"): immp.Nullable(str),
                          immp.Optional("client-updates", False): bool,
                          immp.Optional("session"): immp.Nullable(str)})

    user = immp.Schema({"id": int,
                        immp.Optional("username"): immp.Nullable(str),
                        "first_name": str,
                        immp.Optional("last_name"): immp.Nullable(str)})

    channel = immp.Schema({"id": int,
                           "title": str,
                           "type": "channel",
                           immp.Optional("username"): immp.Nullable(str)})

    entity = immp.Schema({"type": str,
                          "offset": int,
                          "length": int,
                          immp.Optional("url"): immp.Nullable(str),
                          immp.Optional("user"): immp.Nullable(user)})

    _file = {"file_id": str, immp.Optional("file_name"): immp.Nullable(str)}

    _location = {"latitude": float, "longitude": float}

    message = immp.Schema({"message_id": int,
                           "chat": {"id": int},
                           "date": int,
                           immp.Optional("edit_date"): immp.Nullable(int),
                           immp.Optional("from"): immp.Nullable(user),
                           immp.Optional("forward_from"): immp.Nullable(user),
                           immp.Optional("forward_date"): immp.Nullable(int),
                           immp.Optional("forward_from_chat"): immp.Nullable(channel),
                           immp.Optional("forward_from_message_id"): immp.Nullable(int),
                           immp.Optional("forward_signature"): immp.Nullable(str),
                           immp.Optional("text"): immp.Nullable(str),
                           immp.Optional("caption"): immp.Nullable(str),
                           immp.Optional("entities", list): [entity],
                           immp.Optional("caption_entities", list): [entity],
                           immp.Optional("photo", list): [_file],
                           immp.Optional("sticker"): immp.Nullable({immp.Optional("emoji"):
                                                                        immp.Nullable(str),
                                                                    "file_id": str}),
                           immp.Optional("animation"): immp.Nullable(_file),
                           immp.Optional("video"): immp.Nullable(_file),
                           immp.Optional("video_note"): immp.Nullable(_file),
                           immp.Optional("audio"): immp.Nullable(_file),
                           immp.Optional("voice"): immp.Nullable(_file),
                           immp.Optional("document"): immp.Nullable(_file),
                           immp.Optional("location"): immp.Nullable(_location),
                           immp.Optional("venue"): immp.Nullable({"location": _location,
                                                                  "title": str,
                                                                  "address": str}),
                           immp.Optional("poll"): immp.Nullable({"question": str,
                                                                 "is_closed": bool}),
                           immp.Optional("group_chat_created", False): bool,
                           immp.Optional("new_chat_members", list): [user],
                           immp.Optional("left_chat_member"): immp.Nullable(user),
                           immp.Optional("new_chat_title"): immp.Nullable(str),
                           immp.Optional("new_chat_photo", list): [_file],
                           immp.Optional("delete_chat_photo", False): bool,
                           immp.Optional("migrate_to_chat_id"): immp.Nullable(int)})

    # Circular references to embedded messages.
    message.raw.update({immp.Optional("reply_to_message"): immp.Nullable(message),
                        immp.Optional("pinned_message"): immp.Nullable(message)})

    update = immp.Schema({"update_id": int,
                          immp.Optional(immp.Any("message", "edited_message",
                                                 "channel_post", "edited_channel_post")): message})

    def api(result=None):
        success = {"ok": True}
        if result:
            success["result"] = result
        return immp.Schema(immp.Any(success,
                                    {"ok": False,
                                     "description": str,
                                     "error_code": int}))

    me = api(user)

    file = api({"file_path": str})

    send = api(message)

    chat = api({"type": str,
                immp.Optional("title"): immp.Nullable(str)})

    updates = api([update])


class TelegramAPIConnectError(immp.PlugError):
    """
    Generic error whilst attempting to call the Telegram API.
    """


class TelegramAPIRequestError(immp.PlugError):
    """
    Generic error from the Telegram API.
    """


class _HiddenSender(Exception):
    pass


class TelegramUser(immp.User):
    """
    User present in Telegram.
    """

    @classmethod
    def from_bot_user(cls, telegram, json):
        """
        Convert a user :class:`dict` (attached to a message) to a :class:`.User`.

        Args:
            telegram (.TelegramPlug):
                Related plug instance that provides the user.
            json (dict):
                Telegram API `User <https://core.telegram.org/bots/api#user>`_ object.

        Returns:
            .TelegramUser:
                Parsed user object.
        """
        user = _Schema.user(json)
        real_name = user["first_name"]
        if user["last_name"]:
            real_name = "{} {}".format(real_name, user["last_name"])
        avatar = None
        if user["username"]:
            avatar = "https://t.me/i/userpic/320/{}.jpg".format(user["username"])
        return cls(id_=user["id"],
                   plug=telegram,
                   username=user["username"],
                   real_name=real_name,
                   avatar=avatar,
                   raw=user)

    @classmethod
    def from_bot_channel(cls, telegram, json):
        """
        Convert a chat :class:`dict` (attached to a message) to a :class:`.User`.

        Args:
            telegram (.TelegramPlug):
                Related plug instance that provides the user.
            json (dict):
                Telegram API `Chat <https://core.telegram.org/bots/api#chat>`_ object.

        Returns:
            .TelegramUser:
                Parsed user object.
        """
        chat = _Schema.channel(json)
        if chat["id"] == -1001228946795:
            raise _HiddenSender
        return cls(id_=chat["id"],
                   plug=telegram,
                   username=chat["username"],
                   real_name=chat["title"],
                   raw=chat)

    @classmethod
    def from_proto_user(cls, telegram, user):
        """
        Convert a :class:`telethon.tl.types.User` into a :class:`.User`.

        Args:
            telegram (.TelegramPlug):
                Related plug instance that provides the user.
            user (telethon.tl.types.User):
                Telegram user retrieved from the MTProto API.

        Returns:
            .TelegramUser:
                Parsed user object.
        """
        real_name = user.first_name
        if user.last_name:
            real_name = "{} {}".format(real_name, user.last_name)
        avatar = None
        if user.username and user.photo:
            avatar = "https://t.me/i/userpic/320/{}.jpg".format(user.username)
        return cls(id_=user.id,
                   plug=telegram,
                   username=user.username,
                   real_name=real_name,
                   avatar=avatar,
                   raw=user)

    @classmethod
    def from_entity(cls, telegram, entity):
        """
        Convert a client entity row into a :class:`.User`.

        Args:
            telegram (.TelegramPlug):
                Related plug instance that provides the user.
            entity ((str, str, str) tuple):
                ID, username and real name of a cached Telegram user.

        Returns:
            .TelegramUser:
                Parsed user object.
        """
        id_, username, name = entity
        return cls(id_=id_,
                   plug=telegram,
                   username=username,
                   real_name=name,
                   raw=entity)

    @property
    def link(self):
        if self.username:
            return "https://t.me/{}".format(self.username)
        else:
            return "tg://user?id={}".format(self.id)

    @link.setter
    def link(self, value):
        pass


class TelegramSegment(immp.Segment):
    """
    Plug-friendly representation of Telegram message formatting.
    """

    @classmethod
    def to_html(cls, telegram, segment):
        """
        Convert a :class:`.Segment` into HTML suitable for Telegram's automatic parsing.

        Args:
            telegram (.TelegramPlug):
                Related plug instance to cross-reference users.
            segment (.Segment)
                Message segment created by another plug.

        Returns:
            str:
                HTML-formatted string.
        """
        text = segment.text.replace("&", "&amp;").replace("<", "&lt;").replace(">", "&gt;")
        # Any form of tag nesting (e.g. bold inside italic) isn't supported, so at most one type of
        # formatting may apply for each segment.
        if segment.mention and segment.mention.plug.network_id == telegram.network_id:
            if segment.mention.username:
                # Telegram will parse this automatically.
                text = "@{}".format(segment.mention.username)
            else:
                # Make a link that looks like a mention.
                text = ("<a href=\"tg://user?id={}\">{}</a>"
                        .format(segment.mention.id, segment.mention.real_name))
        elif segment.link:
            text = "<a href=\"{}\">{}</a>".format(segment.link, text)
        elif segment.pre:
            text = "<pre>{}</pre>".format(text)
        elif segment.code:
            text = "<code>{}</code>".format(text)
        elif segment.bold:
            text = "<b>{}</b>".format(text)
        elif segment.italic:
            text = "<i>{}</i>".format(text)
        return text


class TelegramRichText(immp.RichText):
    """
    Wrapper for Telegram-specific parsing of formatting.
    """

    @classmethod
    def _from_changes(cls, text, changes):
        segments = []
        points = list(sorted(changes.keys()))
        formatting = {}
        # Iterate through text in change start/end pairs.
        for start, end in zip([0] + points, points + [len(text)]):
            formatting.update(changes[start])
            if start == end:
                # Zero-length segment at the start or end, ignore it.
                continue
            segments.append(immp.Segment(text[start:end], **formatting))
        return cls(segments)

    @classmethod
    async def from_bot_entities(cls, telegram, text, entities):
        """
        Convert a string annotated by Telegram's entities to :class:`.RichText`.

        Args:
            telegram (.TelegramPlug):
                Related plug instance that provides the text.
            text (str):
                Plain text without formatting.
            entities (dict list):
                List of Telegram API `MessageEntity
                <https://core.telegram.org/bots/api#messageentity>`_ objects.

        Returns:
            .TelegramRichText:
                Parsed rich text container.
        """
        if not text:
            return None
        elif not entities:
            return immp.RichText([immp.Segment(text)])
        # Telegram entities assume the text is UTF-16.
        encoded = text.encode("utf-16-le")
        changes = defaultdict(dict)
        for json in entities:
            entity = _Schema.entity(json)
            start = entity["offset"] * 2
            end = start + (entity["length"] * 2)
            if entity["type"] in ("bold", "italic", "code", "pre"):
                key = entity["type"]
                value, clear = True, False
            elif entity["type"] in ("url", "email"):
                key = "link"
                value, clear = encoded[start:end].decode("utf-16-le"), None
            elif entity["type"] == "text_url":
                key = "link"
                value, clear = entity["url"], None
            elif entity["type"] == "mention":
                key = "mention"
                username = encoded[start + 2:end].decode("utf-16-le")
                value, clear = await telegram.user_from_username(username), None
            elif entity["type"] == "text_mention":
                key = "mention"
                value, clear = TelegramUser.from_bot_user(telegram, entity["user"]), None
            else:
                continue
            changes[start][key] = value
            changes[end][key] = clear
        rich = cls._from_changes(encoded, changes)
        for segment in rich:
            segment.text = segment.text.decode("utf-16-le")
        return rich

    @classmethod
    async def from_proto_entities(cls, telegram, text, entities):
        """
        Convert a string annotated by Telegram's entities to :class:`.RichText`.

        Args:
            telegram (.TelegramPlug):
                Related plug instance that provides the text.
            text (str):
                Plain text without formatting.
            entities (telethon.types.TypeMessageEntity list):
                List of Telegram entity objects.

        Returns:
            .TelegramRichText:
                Parsed rich text container.
        """
        if not text:
            return None
        elif not entities:
            return immp.RichText([immp.Segment(text)])
        changes = defaultdict(dict)
        for entity in entities:
            value, clear = True, False
            if isinstance(entity, tl.types.MessageEntityBold):
                key = "bold"
            elif isinstance(entity, tl.types.MessageEntityItalic):
                key = "italic"
            elif isinstance(entity, tl.types.MessageEntityCode):
                key = "code"
            elif isinstance(entity, tl.types.MessageEntityPre):
                key = "pre"
            elif isinstance(entity, (tl.types.MessageEntityEmail, tl.types.MessageEntityUrl)):
                key = "link"
                value, clear = text[entity.offset:entity.offset + entity.length], None
            elif isinstance(entity, tl.types.MessageEntityTextUrl):
                key = "link"
                value, clear = entity.url, None
            elif isinstance(entity, tl.types.MessageEntityMention):
                key = "mention"
                username = text[entity.offset + 1:entity.offset + entity.length]
                value, clear = await telegram.user_from_username(username), None
            elif isinstance(entity, tl.types.MessageEntityMentionName):
                key = "mention"
                value, clear = await telegram.user_from_id(entity.user_id), None
            else:
                continue
            changes[entity.offset][key] = value
            changes[entity.offset + entity.length][key] = clear
        return cls._from_changes(text, changes)


class TelegramFile(immp.File):
    """
    File attachment originating from Telegram.
    """

    @classmethod
    async def from_id(cls, telegram, id_, type_=immp.File.Type.unknown, name=None):
        """
        Generate a file using the bot API URL for a Telegram file.

        Args:
            telegram (.TelegramPlug):
                Related plug instance that provides the file.
            id (str):
                File ID as provided in the bot API, or constructed from a raw MTProto file.
            type (.File.Type):
                Corresponding file type.
            name (str):
                Original filename, if available for the file format.

        Returns:
            .TelegramFile:
                Parsed file object.
        """
        file = await telegram._api("getFile", _Schema.file, params={"file_id": id_})
        url = ("https://api.telegram.org/file/bot{}/{}"
               .format(telegram.config["token"], file["file_path"]))
        return immp.File(name, type_, url)


class TelegramMessage(immp.Message):
    """
    Message originating from Telegram.
    """

    _file_types = ("animation", "video", "video_note", "audio", "voice", "document")

    @classmethod
    async def from_bot_message(cls, telegram, json):
        """
        Convert an API message :class:`dict` to a :class:`.Message`.

        Args:
            telegram (.TelegramPlug):
                Related plug instance that provides the event.
            json (dict):
                Telegram API `message <https://core.telegram.org/bots/api#message>`_ object.

        Returns:
            .TelegramMessage:
                Parsed message object.
        """
        message = _Schema.message(json)
        # Message IDs are just a sequence, only unique to their channel and not the whole network.
        # Pair with the chat ID for a network-unique value.
        id_ = "{}:{}".format(message["chat"]["id"], message["message_id"])
        revision = message["edit_date"] or message["date"]
        at = datetime.fromtimestamp(message["date"], timezone.utc)
        channel = immp.Channel(telegram, message["chat"]["id"])
        edited = bool(message["edit_date"])
        text = await TelegramRichText.from_bot_entities(telegram, message["text"],
                                                        message["entities"])
        user = None
        action = False
        reply_to = None
        joined = None
        left = None
        title = None
        attachments = []
        if message["from"]:
            user = TelegramUser.from_bot_user(telegram, message["from"])
        if message["reply_to_message"]:
            reply_to = await cls.from_bot_message(telegram, message["reply_to_message"])
        # At most one of these fields will be set.
        if message["group_chat_created"]:
            action = True
            text = immp.RichText([immp.Segment("created the group "),
                                  immp.Segment(message["chat"]["title"], bold=True)])
        elif message["new_chat_members"]:
            joined = [(TelegramUser.from_bot_user(telegram, member))
                      for member in message["new_chat_members"]]
            action = True
            if joined == [user]:
                text = "joined group via invite link"
            else:
                text = immp.RichText()
                for join in joined:
                    text.append(immp.Segment(", " if text else "invited "),
                                immp.Segment(join.real_name, bold=True, link=join.link))
        elif message["left_chat_member"]:
            left = [TelegramUser.from_bot_user(telegram, message["left_chat_member"])]
            action = True
            if left == [user]:
                text = "left group"
            else:
                part = left[0]
                text = immp.RichText([immp.Segment("removed "),
                                      immp.Segment(part.real_name, bold=True, link=part.link)])
        elif message["new_chat_title"]:
            title = message["new_chat_title"]
            action = True
            text = immp.RichText([immp.Segment("changed group name to "),
                                  immp.Segment(title, bold=True)])
        elif message["new_chat_photo"]:
            action = True
            text = "changed group photo"
            photo = max(message["new_chat_photo"], key=lambda photo: photo["height"])
            attachments.append(await TelegramFile.from_id(telegram, photo["file_id"],
                                                          immp.File.Type.image))
        elif message["delete_chat_photo"]:
            action = True
            text = "removed group photo"
        elif message["pinned_message"]:
            action = True
            text = "pinned a message"
            attachments.append(await cls.from_bot_message(telegram, message["pinned_message"]))
        elif message["photo"]:
            # This is a list of resolutions, find the original sized one to return.
            photo = max(message["photo"], key=lambda photo: photo["height"])
            attachments.append(await TelegramFile.from_id(telegram, photo["file_id"],
                                                          immp.File.Type.image))
            if message["caption"]:
                text = await TelegramRichText.from_bot_entities(telegram, message["caption"],
                                                                message["caption_entities"])
        elif message["sticker"]:
            attachments.append(await TelegramFile.from_id(telegram, message["sticker"]["file_id"],
                                                          immp.File.Type.image))
            # All real stickers should have an emoji, but webp images uploaded as photos are
            # incorrectly categorised as stickers in the API response.
            if not text and message["sticker"]["emoji"]:
                action = True
                text = "sent {} sticker".format(message["sticker"]["emoji"])
        elif any(message[key] for key in cls._file_types):
            for key in cls._file_types:
                if message[key]:
                    obj = message[key]
                    break
            type_ = immp.File.Type.image if key == "animation" else immp.File.Type.unknown
            attachments.append(await TelegramFile.from_id(telegram, obj["file_id"], type_,
                                                          obj["file_name"]))
        elif message["venue"]:
            attachments.append(immp.Location(latitude=message["venue"]["location"]["latitude"],
                                             longitude=message["venue"]["location"]["longitude"],
                                             name=message["venue"]["title"],
                                             address=message["venue"]["address"]))
        elif message["location"]:
            attachments.append(immp.Location(latitude=message["location"]["latitude"],
                                             longitude=message["location"]["longitude"]))
        elif message["poll"]:
            action = True
            prefix = "closed the" if message["poll"]["is_closed"] else "opened a"
            text = immp.RichText([immp.Segment("{} poll: ".format(prefix)),
                                  immp.Segment(message["poll"]["question"], bold=True)])
        elif not text:
            # No support for this message type.
            raise NotImplementedError
        common = dict(id_=id_,
                      revision=revision,
                      at=at,
                      channel=channel,
                      edited=edited,
                      user=user,
                      raw=message)
        if message["forward_date"]:
            # Event is a message containing another message.  Forwarded messages have no ID, so we
            # use a Message instead of a SentMessage here, unless they come from a channel.
            forward_id = forward_channel = forward_user = None
            if message["forward_from"]:
                forward_user = TelegramUser.from_bot_user(telegram, message["forward_from"])
            elif message["forward_from_chat"]:
                forward_channel = immp.Channel(telegram, message["forward_from_chat"]["id"])
                try:
                    forward_user = TelegramUser.from_bot_channel(telegram,
                                                                 message["forward_from_chat"])
                except _HiddenSender:
                    if message["forward_signature"]:
                        forward_user = immp.User(real_name=message["forward_signature"])
                if message["forward_from_message_id"]:
                    forward_id = "{}:{}".format(message["forward_from_chat"]["id"],
                                                message["forward_from_message_id"])
            forward_common = dict(text=text,
                                  user=forward_user,
                                  action=action,
                                  reply_to=reply_to,
                                  joined=joined,
                                  left=left,
                                  title=title,
                                  attachments=attachments,
                                  raw=message)
            if forward_id:
                forward = immp.SentMessage(id_=forward_id,
                                           channel=forward_channel,
                                           **forward_common)
            else:
                forward = immp.Message(**forward_common)
            # Embed the inner message as an attachment.
            return immp.SentMessage(attachments=[forward], **common)
        else:
            return immp.SentMessage(text=text,
                                    action=action,
                                    reply_to=reply_to,
                                    joined=joined,
                                    left=left,
                                    title=title,
                                    attachments=attachments,
                                    **common)

    @classmethod
    async def from_bot_update(cls, telegram, update):
        """
        Convert an API update :class:`dict` to a :class:`.Message`.

        Args:
            telegram (.TelegramPlug):
                Related plug instance that provides the event.
            update (dict):
                Telegram API `update <https://core.telegram.org/bots/api#update>`_ object.

        Returns:
            .TelegramMessage:
                Parsed message object.
        """
        for key in ("message", "channel_post"):
            if update.get(key):
                return await cls.from_bot_message(telegram, update[key])
            elif update.get("edited_{}".format(key)):
                return await cls.from_bot_message(telegram, update["edited_{}".format(key)])

    @classmethod
    async def from_proto_message(cls, telegram, message):
        """
        Convert a Telegram message event to a :class:`.Message`.

        Args:
            telegram (.TelegramPlug):
                Related plug instance that provides the event.
            message (telethon.tl.custom.Message):
                Received message from an event or get request.

        Returns:
            .TelegramMessage:
                Parsed message object.
        """
        id_ = "{}:{}".format(message.chat_id, message.id)
        edited = bool(message.edit_date)
        if edited:
            revision = int(message.edit_date.timestamp())
        elif message.date:
            revision = int(message.date.timestamp())
        else:
            revision = None
        text = await TelegramRichText.from_proto_entities(telegram, message.message,
                                                          message.entities)
        user = TelegramUser.from_proto_user(telegram, await message.get_sender())
        action = False
        reply_to = None
        joined = []
        left = []
        title = None
        attachments = []
        if message.reply_to_msg_id:
            reply_to = await telegram.get_message(message.reply_to_msg_id)
        if message.photo:
            try:
                attach = await TelegramFile.from_id(telegram, pack_bot_file_id(message.photo),
                                                    immp.File.Type.image)
            except TelegramAPIRequestError as e:
                log.warning("Unable to fetch attachment", exc_info=e)
            else:
                attachments.append(attach)
        elif message.document:
            type_ = immp.File.Type.unknown
            name = None
            for attr in message.document.attributes:
                if isinstance(attr, tl.types.DocumentAttributeSticker):
                    type_ = immp.File.Type.image
                    if attr.alt and not text:
                        text = "sent {} sticker".format(attr.alt)
                        action = True
                elif isinstance(attr, tl.types.DocumentAttributeAnimated):
                    type_ = immp.File.Type.image
                elif isinstance(attr, tl.types.DocumentAttributeFilename):
                    name = attr.file_name
            try:
                attach = await TelegramFile.from_id(telegram, pack_bot_file_id(message.document),
                                                    type_, name)
            except TelegramAPIRequestError as e:
                log.warning("Unable to fetch attachment", exc_info=e)
            else:
                attachments.append(attach)
        elif message.poll:
            action = True
            prefix = "closed the" if message.poll.poll.closed else "opened a"
            text = immp.RichText([immp.Segment("{} poll: ".format(prefix)),
                                  immp.Segment(message.poll.poll.question, bold=True)])
        if message.action:
            action = True
            if isinstance(message.action, tl.types.MessageActionChatCreate):
                text = immp.RichText([immp.Segment("created the group "),
                                      immp.Segment(message.action.title, bold=True)])
            elif isinstance(message.action, tl.types.MessageActionChatJoinedByLink):
                joined = [user]
                text = "joined group via invite link"
            elif isinstance(message.action, tl.types.MessageActionChatAddUser):
                joined = await gather(*(telegram.user_from_id(id_) for id_ in message.action.users))
                if joined == [user]:
                    text = "joined group"
                else:
                    text = immp.RichText()
                    for join in joined:
                        text.append(immp.Segment(", " if text else "invited "),
                                    immp.Segment(join.real_name, link=join.link))
            elif isinstance(message.action, tl.types.MessageActionChatDeleteUser):
                left = [await telegram.user_from_id(message.action.user_id)]
                if left == [user]:
                    text = "left group"
                else:
                    part = left[0]
                    text = immp.RichText([immp.Segment("removed "),
                                          immp.Segment(part.real_name, bold=True, link=part.link)])
            elif isinstance(message.action, tl.types.MessageActionChatEditTitle):
                title = message.action.title
                text = immp.RichText([immp.Segment("changed group name to "),
                                      immp.Segment(title, bold=True)])
            elif isinstance(message.action, tl.types.MessageActionChatEditPhoto):
                text = "changed group photo"
            elif isinstance(message.action, tl.types.MessageActionChatDeletePhoto):
                text = "removed group photo"
            elif isinstance(message.action, tl.types.MessageActionPinMessage):
                attachments.append(reply_to)
                reply_to = None
                text = "pinned message"
            else:
                raise NotImplementedError
        if not text and not attachments:
            # No support for this message type.
            raise NotImplementedError
        common = dict(id_=id_,
                      revision=revision,
                      at=message.date,
                      channel=immp.Channel(telegram, message.chat_id),
                      edited=edited,
                      user=user,
                      raw=message)
        if message.forward:
            # Event is a message containing another message.  Forwarded messages have no ID, so we
            # use a Message instead of a SentMessage here, unless they come from a channel.
            forward_id = forward_channel = forward_user = None
            if message.forward.channel_id and message.forward.channel_post:
                forward_channel = immp.Channel(telegram, message.forward.chat_id)
                forward_id = "{}:{}".format(message.forward.chat_id,
                                            message.forward.channel_post)
                if message.forward.post_author:
                    forward_user = immp.User(real_name=message.forward.post_author)
                else:
                    chat = await message.forward.get_chat()
                    forward_user = immp.User(real_name=chat.title)
            elif message.forward.sender_id:
                forward_user = TelegramUser.from_proto_user(telegram,
                                                            await message.forward.get_sender())
            elif message.forward.from_name:
                forward_user = immp.User(real_name=message.forward.from_name)
            forward_common = dict(text=text,
                                  user=forward_user,
                                  action=action,
                                  reply_to=reply_to,
                                  joined=joined,
                                  left=left,
                                  title=title,
                                  attachments=attachments,
                                  raw=message)
            if forward_id:
                forward = immp.SentMessage(id_=forward_id,
                                           channel=forward_channel,
                                           **forward_common)
            else:
                forward = immp.Message(**forward_common)
            # Embed the inner message as an attachment.
            return immp.SentMessage(attachments=[forward], **common)
        else:
            return immp.SentMessage(text=text,
                                    action=action,
                                    reply_to=reply_to,
                                    joined=joined,
                                    left=left,
                                    title=title,
                                    attachments=attachments,
                                    **common)


if SQLiteSession:

    class Session(SQLiteSession):

        def _execute_multi(self, statement, *values):
            cursor = self._cursor()
            try:
                return cursor.execute(statement, values).fetchall()
            finally:
                cursor.close()

        def get_user_entities(self):
            return self._execute_multi("SELECT id, username, name FROM entities WHERE id > 0")

        def get_chat_entities(self):
            return self._execute_multi("SELECT id, username, name FROM entities WHERE id < 0")

        def get_entity(self, id_):
            return self._execute("SELECT id, username, name FROM entities WHERE id = ?", id_)


class TelegramPlug(immp.HTTPOpenable, immp.Plug):
    """
    Plug for a `Telegram <https://telegram.org>`_ bot.
    """

    schema = _Schema.config

    network_name = "Telegram"

    @property
    def network_id(self):
        return "telegram:{}".format(self._bot_user["id"]) if self._bot_user else None

    def __init__(self, name, config, host):
        super().__init__(name, config, host)
        if bool(self.config["api-id"]) != bool(self.config["api-hash"]):
            raise immp.ConfigError("Both of API ID and hash must be given")
        if self.config["client-updates"] and not self.config["api-id"]:
            raise immp.ConfigError("Client updates require API ID and hash")
        if self.config["session"] and not self.config["api-id"]:
            raise immp.ConfigError("Session file requires API ID and hash")
        # Connection objects that need to be closed on disconnect.
        self._bot_user = self._receive = self._client = None
        self._closing = False
        # Temporary tracking of migrated chats for the current session.
        self._migrations = {}
        # Update ID from which to retrieve the next batch.  Should be one higher than the max seen.
        self._offset = 0

    async def _api(self, endpoint, schema=_Schema.api(), **kwargs):
        url = "https://api.telegram.org/bot{}/{}".format(self.config["token"], endpoint)
        log.debug("Making API request to %r", endpoint)
        try:
            async with self.session.post(url, **kwargs) as resp:
                try:
                    json = await resp.json()
                    data = schema(json)
                except ClientResponseError as e:
                    raise TelegramAPIConnectError("Bad response with code: {}"
                                                  .format(resp.status)) from e
        except ClientError as e:
            raise TelegramAPIConnectError("Request failed") from e
        except TimeoutError as e:
            raise TelegramAPIConnectError("Request timed out") from e
        if not data["ok"]:
            raise TelegramAPIRequestError(data["error_code"], data["description"])
        return data["result"]

    async def start(self):
        await super().start()
        self._closing = False
        self._bot_user = await self._api("getMe", _Schema.me)
        if self.config["api-id"] and self.config["api-hash"]:
            if not TelegramClient:
                raise immp.ConfigError("API ID/hash specified but Telethon is not installed")
            log.debug("Starting client")
            self._client = TelegramClient(Session(self.config["session"]),
                                          self.config["api-id"], self.config["api-hash"])
        if self._client and self.config["client-updates"]:
            log.debug("Adding client event handlers")
            self._client.add_event_handler(self._handle_raw)
            for event in (events.NewMessage, events.MessageEdited, events.ChatAction):
                self._client.add_event_handler(self._handle, event)
        else:
            log.debug("Starting update long-poll")
            self._receive = ensure_future(self._poll())
        if self._client:
            await self._client.start(bot_token=self.config["token"])

    async def stop(self):
        await super().stop()
        self._closing = True
        if self._receive:
            log.debug("Stopping update long-poll")
            self._receive.cancel()
            self._receive = None
        if self._client:
            log.debug("Closing client")
            await self._client.disconnect()
            self._client = None
        self._bot_user = None
        self._offset = 0
        if self._migrations:
            log.warning("Chat migrations require a config update before next run")

    async def user_from_id(self, id_):
        if not self._client:
            log.debug("Client auth required to look up users")
            return None
        try:
            data = await self._client(tl.functions.users.GetFullUserRequest(int(id_)))
        except BadRequestError:
            entity = self._client.session.get_entity(id_)
            return TelegramUser.from_entity(self, entity) if entity else None
        else:
            return TelegramUser.from_proto_user(self, data.user)

    async def user_from_username(self, username):
        if not self._client:
            log.debug("Client auth required to look up users")
            return None
        try:
            data = await self._client(tl.functions.contacts.ResolveUsernameRequest(username))
        except BadRequestError:
            return None
        else:
            return TelegramUser.from_proto_user(self, data.users[0]) if data.users else None

    async def user_is_system(self, user):
        return user.id == str(self._bot_user["id"])

    async def public_channels(self):
        if not self._client:
            log.debug("Client auth required to look up channels")
            return None
        # Use the session cache to find all "seen" chats -- not guaranteed to be a complete list.
        return [immp.Channel(self, chat[0]) for chat in self._client.session.get_chat_entities()]

    async def private_channels(self):
        if not self._client:
            log.debug("Client auth required to look up channels")
            return None
        # Private channels just use user IDs, so return all users we know about.  Note that these
        # channels aren't usable unless the user has messaged the bot first.
        return [immp.Channel(self, chat[0]) for chat in self._client.session.get_user_entities()]

    async def channel_for_user(self, user):
        if not isinstance(user, TelegramUser):
            return None
        try:
            await self._api("getChat", params={"chat_id": user.id})
        except TelegramAPIRequestError as e:
            log.warning("Failed to retrieve user %s channel", user.id, exc_info=e)
            # Can't create private channels, users must initiate conversations with bots.
            return None
        else:
            return immp.Channel(self, user.id)

    async def channel_is_private(self, channel):
        return int(channel.source) > 0

    async def channel_title(self, channel):
        if await channel.is_private():
            return None
        if self._client:
            row = self._client.session.get_entity(channel.source)
            if row and row[2]:
                return row[2]
        try:
            data = await self._api("getChat", _Schema.chat, params={"chat_id": channel.source})
        except TelegramAPIRequestError as e:
            log.warning("Failed to retrieve channel %s title", channel.source, exc_info=e)
            return None
        else:
            return data["title"]

    async def channel_rename(self, channel, title):
        await self._api("setChatTitle", params={"chat_id": channel.source, "title": title})

    async def channel_members(self, channel):
        if not self._client:
            log.debug("Client auth required to list channel members")
            return None
        # Channel and supergroup chat IDs have a bot-API-only prefix to distinguish them.
        if channel.source.startswith("-100"):
            chat = int(channel.source[4:])
            users = []
            try:
                while True:
                    data = await self._client(tl.functions.channels.GetParticipantsRequest(
                        chat, tl.types.ChannelParticipantsRecent(), len(users), 1000, 0))
                    if data.users:
                        users += [TelegramUser.from_proto_user(self, user) for user in data.users]
                    else:
                        break
            except BadRequestError:
                return None
            else:
                return users
        else:
            chat = abs(int(channel.source))
            try:
                data = await self._client(tl.functions.messages.GetFullChatRequest(chat))
            except BadRequestError:
                # Private channels should just contain the bot and the corresponding user.
                if channel.source == str(self._bot_user["id"]):
                    return [TelegramUser.from_bot_user(self, self._bot_user)]
                elif int(channel.source) > 0:
                    entity = self._client.session.get_entity(channel.source)
                    if entity:
                        return [TelegramUser.from_bot_user(self, self._bot_user),
                                await self.user_from_id(channel.source)]
                    else:
                        return None
            else:
                return [TelegramUser.from_proto_user(self, user) for user in data.users]

    async def channel_remove(self, channel, user):
        await self._api("kickChatMember", params={"chat_id": channel.source, "user_id": user.id})

    async def channel_history(self, channel, before=None):
        if not self._client:
            log.debug("Client auth required to retrieve messages")
            return []
        elif not before:
            log.debug("Before message required to retrieve messages")
            return []
        chat, message = (int(field) for field in before.id.split(":", 1))
        ids = list(range(max(message - 50, 1), message))
        history = filter(None, await self._client.get_messages(entity=chat, ids=ids))
        tasks = (TelegramMessage.from_proto_message(self, message) for message in history)
        results = await gather(*tasks, return_exceptions=True)
        messages = []
        for result in results:
            if isinstance(result, NotImplementedError):
                continue
            elif isinstance(result, Exception):
                raise result
            else:
                messages.append(result)
        return messages

    async def get_message(self, id_):
        if not self._client:
            log.debug("Client auth required to retrieve messages")
            return None
        message = await self._client.get_messages(None, ids=id_)
        if not message:
            return None
        try:
            return await TelegramMessage.from_proto_message(self, message)
        except NotImplementedError:
            return None

    async def _form_data(self, base, field, attach):
        data = FormData(base)
        if attach.source:
            data.add_field(field, attach.source)
        else:
            img_resp = await attach.get_content(self.session)
            data.add_field(field, img_resp.content, filename=attach.title or field)
        return data

    async def _upload_attachment(self, chat, msg, attach):
        # Upload an image file to Telegram in its own message.
        # Prefer a source URL if available, else fall back to re-uploading the file.
        base = {"chat_id": str(chat)}
        if msg.user:
            rich = immp.RichText([immp.Segment(msg.user.real_name or msg.user.username,
                                               bold=True, italic=True),
                                  immp.Segment(" sent an image", italic=True)])
            text = "".join(TelegramSegment.to_html(self, segment) for segment in rich)
            base["caption"] = text
            base["parse_mode"] = "HTML"
        if attach.type == immp.File.Type.image:
            data = await self._form_data(base, "photo", attach)
            try:
                return await self._api("sendPhoto", _Schema.send, data=data)
            except (TelegramAPIConnectError, TelegramAPIRequestError):
                log.debug("Failed to upload image, falling back to document upload")
        data = await self._form_data(base, "document", attach)
        try:
            return await self._api("sendDocument", _Schema.send, data=data)
        except TelegramAPIConnectError as e:
            log.warning("Failed to upload file", exc_info=e)
            return None

    def _requests(self, chat, msg):
        requests = []
        if msg.text or msg.reply_to:
            quote = False
            reply_to = ""
            if isinstance(msg.reply_to, immp.Receipt):
                # Reply natively to the given parent message.
                reply_to = int(msg.reply_to.id.split(":")[1])
            elif isinstance(msg.reply_to, immp.Message):
                quote = True
            edited = msg.edited if isinstance(msg, immp.Receipt) else False
            rich = msg.render(edit=edited, quote_reply=quote)
            text = "".join(TelegramSegment.to_html(self, segment) for segment in rich)
            requests.append(self._api("sendMessage", _Schema.send,
                                      params={"chat_id": chat,
                                              "text": text,
                                              "parse_mode": "HTML",
                                              "reply_to_message_id": reply_to}))
        for attach in msg.attachments:
            if isinstance(attach, immp.File):
                requests.append(self._upload_attachment(chat, msg, attach))
            elif isinstance(attach, immp.Location):
                requests.append(self._api("sendLocation", _Schema.send,
                                          params={"chat_id": chat,
                                                  "latitude": str(attach.latitude),
                                                  "longitude": str(attach.longitude)}))
                if msg.user:
                    caption = immp.Message(user=msg.user, text="sent a location", action=True)
                    text = "".join(TelegramSegment.to_html(self, segment)
                                   for segment in caption.render())
                    requests.append(self._api("sendMessage", _Schema.send,
                                              params={"chat_id": chat,
                                                      "text": text,
                                                      "parse_mode": "HTML"}))
        return requests

    async def put(self, channel, msg):
        chat = channel.source
        while chat in self._migrations:
            log.debug("Following chat migration: %r -> %r", chat, self._migrations[chat])
            chat = self._migrations[chat]
        requests = []
        for attach in msg.attachments:
            # Generate requests for attached messages first.
            if isinstance(attach, immp.Receipt):
                # Forward the messages natively using the given chat/ID.
                forward_chat, forward_id = map(int, attach.id.split(":"))
                requests.append(self._api("forwardMessage", _Schema.send,
                                          params={"chat_id": chat,
                                                  "from_chat_id": forward_chat,
                                                  "message_id": forward_id}))
            elif isinstance(attach, immp.Message):
                requests += self._requests(chat, attach)
        own_requests = self._requests(chat, msg)
        if requests and not own_requests:
            # Forwarding a message but no content to show who forwarded it.
            info = immp.Message(user=msg.user, action=True, text="forwarded a message")
            own_requests = self._requests(chat, info)
        requests += own_requests
        ids = []
        for request in requests:
            result = await request
            if not result:
                continue
            sent = await TelegramMessage.from_bot_message(self, result)
            self.queue(sent)
            ids.append(sent.id)
        return ids

    async def delete(self, sent):
        chat, message = sent.id.split(":", 1)
        await self._api("deleteMessage", params={"chat_id": chat, "message_id": message})

    def _migrate(self, old, new):
        log.warning("Chat has migrated: %r -> %r", old, new)
        self._migrations[old] = new
        for name, channel in self.host.channels.items():
            if channel.plug is self and channel.source == old:
                log.debug("Updating named channel %r in place", name)
                channel.source = new

    async def _poll(self):
        while not self._closing:
            params = {"offset": self._offset,
                      "timeout": 240}
            fetch = ensure_future(self._api("getUpdates", _Schema.updates, params=params))
            try:
                result = await fetch
            except CancelledError:
                log.debug("Cancelling polling")
                return
            except (TelegramAPIConnectError, TelegramAPIRequestError) as e:
                log.debug("Unexpected response or timeout: %r", e)
                log.debug("Reconnecting in 3 seconds")
                await sleep(3)
                continue
            except Exception as e:
                log.exception("Uncaught exception during long-poll: %r", e)
                raise
            for update in result:
                log.debug("Received an update")
                if "message" in update and update["message"]["migrate_to_chat_id"]:
                    old = update["message"]["chat"]["id"]
                    new = update["message"]["migrate_to_chat_id"]
                    self._migrate(old, new)
                if any(key in update or "edited_{}".format(key) in update
                       for key in ("message", "channel_post")):
                    try:
                        sent = await TelegramMessage.from_bot_update(self, update)
                    except NotImplementedError:
                        log.debug("Skipping message with no usable parts")
                    except CancelledError:
                        log.debug("Cancel request for plug %r getter", self.name)
                        return
                    else:
                        self.queue(sent)
                else:
                    log.debug("Ignoring update with unknown keys: %s", ", ".join(update.keys()))
                self._offset = max(update["update_id"] + 1, self._offset)

    async def _handle_raw(self, event):
        log.debug("Received a %s event", event.__class__.__qualname__)
        if isinstance(event, tl.types.UpdateNewMessage):
            if isinstance(event.message.action, tl.types.MessageActionChatMigrateTo):
                old = event.message.chat_id
                new = int("-100{}".format(event.message.action.channel_id))
                self._migrate(old, new)

    async def _handle(self, event):
        if isinstance(event, events.ChatAction.Event):
            message = event.action_message
        else:
            message = event.message
        try:
            sent = await TelegramMessage.from_proto_message(self, message)
        except NotImplementedError:
            log.debug("Skipping message with no usable parts")
        else:
            self.queue(sent)



"""
Provider of database access to other hooks.

Config:
    url (str):
        Database connection string, as defined by Peewee's :ref:`db_url`.

This hook provides persistent storage via a database.  Any database types supported by Peewee can
be used, though the usual caveats apply: if a hook requires fields specific to a single database
type, the app is effectively locked-in to that type.

Hooks should subclass :class:`.BaseModel` for their data structures.  At startup, they can obtain
the database connection via :attr:`host.resources[DatabaseHook].db`, and use it to create their
database tables via :meth:`peewee.Database.create_tables`.

.. note::
    This hook requires the `Peewee <http://docs.peewee-orm.com>`_ Python module, along with any
    database-specific libraries (e.g. Psycopg2 for PostgreSQL).
"""

import logging

from peewee import Model, Proxy
from playhouse.db_url import connect

import immp


log = logging.getLogger(__name__)


class BaseModel(Model):
    """
    Template model to be used by other hooks.
    """

    class Meta:
        database = Proxy()


class DatabaseHook(immp.ResourceHook):
    """
    Hook that provides generic database access to other hooks, backed by :mod:`peewee`.  Because
    models are in the global scope, they can only be attached to a single database, therefore this
    hook acts as the single source of truth for obtaining a "global" database.

    Attributes:
        db (peewee.Database):
            Connected database instance.
    """

    schema = immp.Schema({"url": str})

    def __init__(self, name, config, host):
        super().__init__(name, config, host)
        self.db = None

    async def start(self):
        log.debug("Opening connection to database")
        self.db = connect(self.config["url"])
        BaseModel._meta.database.initialize(self.db)

    async def stop(self):
        if self.db:
            log.debug("Closing connection to database")
            self.db.close()
            self.db = None



class ConfigError(Exception):
    """
    Error for invalid configuration in a given context.
    """


class PlugError(Exception):
    """
    Error for plug-specific problems.
    """


class HookError(Exception):
    """
    Error for hook-specific problems.
    """



"""
Basic identity management for users in different networks.

Config:
    instance (int):
        Unique instance code.
    plugs (str list):
        List of plug names to accept identities for.
    multiple (bool):
        ``True`` (default) to allow linking multiple accounts from the same network.

Commands:
    id-add <name> <pwd>:
        Create a new identity, or link to an existing one from a second user.
    id-rename <name>:
        Rename the current identity.
    id-password <pwd>:
        Update the password for the current identity.
    id-reset:
        Delete the current identity and all linked users.
    id-role <name> [role]:
        List roles assigned to an identity, or add/remove a given role.

This is a local implementation of the identity protocol, providing self-serve identity linking to
users where a user management backend doesn't otherwise exist.

In order to support multiple copies of this hook with overlapping plugs (e.g. including a private
network in some groups), each hook has an instance code.  If a code isn't defined in the config, a
new one will be assigned at startup.  If multiple hooks are in use, it's important to define these
yourself, so that identities remained assigned to the correct instance.

.. note::
    This hook requires an active :class:`.DatabaseHook` to store data.
"""

from asyncio import gather
from hashlib import sha256
import logging

from peewee import CharField, ForeignKeyField, IntegerField

import immp
from immp.hook.access import AccessPredicate
from immp.hook.command import CommandRole, CommandScope, command
from immp.hook.database import BaseModel, DatabaseHook
from immp.hook.identity import Identity, IdentityProvider


CROSS = "\N{CROSS MARK}"
TICK = "\N{WHITE HEAVY CHECK MARK}"


log = logging.getLogger(__name__)


class IdentityGroup(BaseModel):
    """
    Representation of a single identity.

    Attributes:
        instance (int):
            :class:`IdentityHook` instance code.
        name (str):
            Unique display name.
        pwd (str):
            Hashed password, used by the user to authenticate when linking identities.
        links (.IdentityLink iterable):
            All links contained by this group.
    """

    instance = IntegerField()
    name = CharField()
    pwd = CharField()

    class Meta:
        # Uniqueness constraint for each name in each identity instance.
        indexes = ((("instance", "name"), True),)

    @classmethod
    def hash(cls, pwd):
        return sha256(pwd.encode("utf-8")).hexdigest()

    @classmethod
    def select_links(cls):
        return cls.select(cls, IdentityLink).join(IdentityLink)

    async def to_identity(self, host):
        tasks = []
        plugs = {plug.network_id: plug for plug in host.plugs.values()}
        for link in self.links:
            tasks.append(plugs[link.network].user_from_id(link.user))
        users = await gather(*tasks)
        roles = [role.role for role in self.roles]
        return Identity(self.name, users, roles)

    def __repr__(self):
        return "<{}: #{} {}>".format(self.__class__.__name__, self.id, repr(self.name))


class IdentityLink(BaseModel):
    """
    Single link between an identity and a user.

    Attributes:
        group (.IdentityGroup):
            Containing group instance.
        network (str):
            Network identifier that the user belongs to.
        user (str):
            User identifier as given by the plug.
    """

    group = ForeignKeyField(IdentityGroup, backref="links", on_delete="cascade")
    network = CharField()
    user = CharField()

    def __repr__(self):
        return "<{}: #{} {} @ {} {}>".format(self.__class__.__name__, self.id, repr(self.user),
                                             repr(self.network), repr(self.group))


class IdentityRole(BaseModel):
    """
    Assignment of a role to an identity.

    Attributes:
        group (.IdentityGroup):
            Containing group instance.
        role (str):
            Plain role identifier.
    """

    group = ForeignKeyField(IdentityGroup, backref="roles", on_delete="cascade")
    role = CharField()

    def __repr__(self):
        return "<{}: #{} {} {}>".format(self.__class__.__name__, self.id, repr(self.role),
                                        repr(self.group))


class LocalIdentityHook(immp.Hook, AccessPredicate, IdentityProvider):
    """
    Hook for managing physical users with multiple logical links across different plugs.  This
    effectively provides self-service identities, as opposed to being provided externally.
    """

    schema = immp.Schema({immp.Optional("instance"): immp.Nullable(int),
                          "plugs": [str],
                          immp.Optional("multiple", True): bool})

    _plugs = immp.ConfigProperty([immp.Plug])

    def __init__(self, name, config, host):
        super().__init__(name, config, host)
        self.db = None

    async def start(self):
        if not self.config["instance"]:
            # Find a non-conflicting number and assign it.
            codes = {hook.config["instance"] for hook in self.host.hooks.values()
                     if isinstance(hook, self.__class__)}
            code = 1
            while code in codes:
                code += 1
            log.debug("Assigning instance code %d to hook %r", code, self.name)
            self.config["instance"] = code
        self.db = self.host.resources[DatabaseHook].db
        self.db.create_tables([IdentityGroup, IdentityLink, IdentityRole], safe=True)

    def get(self, name):
        """
        Retrieve the identity group using the given name.

        Args:
            name (str):
                Existing name to query.

        Returns:
            .IdentityGroup:
                Linked identity, or ``None`` if not linked.
        """
        try:
            return IdentityGroup.select_links().where(IdentityGroup.name == name).get()
        except IdentityGroup.DoesNotExist:
            return None

    def find(self, user):
        """
        Retrieve the identity that contains the given user, if one exists.

        Args:
            user (.User):
                Existing user to query.

        Returns:
            .IdentityGroup:
                Linked identity, or ``None`` if not linked.
        """
        if not user or user.plug not in self._plugs:
            return None
        try:
            return (IdentityGroup.select_links()
                                 .where(IdentityGroup.instance == self.config["instance"],
                                        IdentityLink.network == user.plug.network_id,
                                        IdentityLink.user == user.id).get())
        except IdentityGroup.DoesNotExist:
            return None

    async def channel_access(self, channel, user):
        return bool(IdentityLink.get(network=user.plug.network_id, user=user.id))

    async def identity_from_name(self, name):
        group = self.get(name)
        return await group.to_identity(self.host) if group else None

    async def identity_from_user(self, user):
        group = self.find(user)
        return await group.to_identity(self.host) if group else None

    def _test(self, channel, user):
        return channel.plug in self._plugs

    @command("id-add", scope=CommandScope.private, test=_test)
    async def add(self, msg, name, pwd):
        """
        Create a new identity, or link to an existing one from a second user.
        """
        if not msg.user or msg.user.plug not in self._plugs:
            return
        if self.find(msg.user):
            text = "{} Already identified".format(CROSS)
        else:
            pwd = IdentityGroup.hash(pwd)
            exists = False
            try:
                group = IdentityGroup.get(instance=self.config["instance"], name=name)
                exists = True
            except IdentityGroup.DoesNotExist:
                group = IdentityGroup.create(instance=self.config["instance"], name=name, pwd=pwd)
            if exists and not group.pwd == pwd:
                text = "{} Password incorrect".format(CROSS)
            elif not self.config["multiple"] and any(link.network == msg.user.plug.network_id
                                                     for link in group.links):
                text = "{} Already identified on {}".format(CROSS, msg.user.plug.network_name)
            else:
                IdentityLink.create(group=group, network=msg.user.plug.network_id,
                                    user=msg.user.id)
                text = "{} {}".format(TICK, "Added" if exists else "Claimed")
        await msg.channel.send(immp.Message(text=text))

    @command("id-rename", scope=CommandScope.private, test=_test)
    async def rename(self, msg, name):
        """
        Rename the current identity.
        """
        if not msg.user:
            return
        group = self.find(msg.user)
        if not group:
            text = "{} Not identified".format(CROSS)
        elif group.name == name:
            text = "{} No change".format(TICK)
        elif IdentityGroup.select().where(IdentityGroup.instance == self.config["instance"],
                                          IdentityGroup.name == name).exists():
            text = "{} Name already in use".format(CROSS)
        else:
            group.name = name
            group.save()
            text = "{} Claimed".format(TICK)
        await msg.channel.send(immp.Message(text=text))

    @command("id-password", scope=CommandScope.private, test=_test)
    async def password(self, msg, pwd):
        """
        Update the password for the current identity.
        """
        if not msg.user:
            return
        group = self.find(msg.user)
        if not group:
            text = "{} Not identified".format(CROSS)
        else:
            group.pwd = IdentityGroup.hash(pwd)
            group.save()
            text = "{} Changed".format(TICK)
        await msg.channel.send(immp.Message(text=text))

    @command("id-reset", scope=CommandScope.private, test=_test)
    async def reset(self, msg):
        """
        Delete the current identity and all linked users.
        """
        if not msg.user:
            return
        group = self.find(msg.user)
        if not group:
            text = "{} Not identified".format(CROSS)
        else:
            group.delete_instance()
            text = "{} Reset".format(TICK)
        await msg.channel.send(immp.Message(text=text))

    @command("id-role", scope=CommandScope.private, role=CommandRole.admin, test=_test)
    async def role(self, msg, name, role=None):
        """
        List roles assigned to an identity, or add/remove a given role.
        """
        try:
            group = IdentityGroup.get(instance=self.config["instance"], name=name)
        except IdentityGroup.DoesNotExist:
            text = "{} Name not registered".format(CROSS)
        else:
            if role:
                count = IdentityRole.delete().where(IdentityRole.group == group,
                                                    IdentityRole.role == role).execute()
                if count:
                    text = "{} Removed".format(TICK)
                else:
                    IdentityRole.create(group=group, role=role)
                    text = "{} Added".format(TICK)
            else:
                roles = IdentityRole.select().where(IdentityRole.group == group)
                if roles:
                    labels = [role.role for role in roles]
                    text = "Roles for {}: {}".format(name, ", ".join(labels))
                else:
                    text = "No roles for {}.".format(name)
        await msg.channel.send(immp.Message(text=text))



"""
Connect to `Discord <https://discordapp.com>`_ as a bot.

Config:
    token (str):
        Discord token for the bot user.
    bot (bool):
        Whether the token represents a bot user (true by default).
    webhooks ((str, str) dict):
        Mapping from Discord channel IDs to webhook URLs, needed for custom message author
        names and avatars.
    playing (str):
        Optional game activity message to show as the bot's presence.

Note that the token is neither a client ID nor client secret -- you need to enable bot features for
your app, and collect the token from there.  New apps can be created from the `My Apps
<https://discordapp.com/developers/applications/me>`_ page in the developer docs.

Because gateway connections can't customise the sender when pushing new messages, you may also want
an `incoming webhook <https://discordapp.com/developers/docs/resources/webhook>`_  configured for
each channel you intend to send messages to.  A new webhook can be created over the API, or in the
UI via Edit Channel > Webhooks.  A fallback style incorporating the user's name in the message text
will be used in lieu of a webhook, e.g. with direct messages.

.. note::
    This plug requires the `discord.py <https://discordpy.readthedocs.io>`_ Python module.
"""

from asyncio import Condition, ensure_future
from collections import defaultdict
from datetime import timezone
from functools import partial
from io import BytesIO
import logging
import re
from textwrap import wrap

import discord
from emoji import emojize

import immp


log = logging.getLogger(__name__)


class _Schema:

    config = immp.Schema({"token": str,
                          immp.Optional("bot", True): bool,
                          immp.Optional("webhooks", dict): {str: str},
                          immp.Optional("playing"): immp.Nullable(str)})

    webhook = immp.Schema(immp.Any({"code": int, "message": str}, {"id": str}))


class DiscordAPIError(immp.PlugError):
    """
    Generic error from the Discord API.
    """


class DiscordUser(immp.User):
    """
    User present in Discord.
    """

    @classmethod
    def from_user(cls, discord, user):
        """
        Convert a :class:`discord.User` into a :class:`.User`.

        Args:
            discord (.DiscordPlug):
                Related plug instance that provides the user.
            user (discord.User):
                Hangups user object retrieved from the user list.

        Returns:
            .DiscordUser:
                Parsed user object.
        """
        username = "{}#{}".format(user.name, user.discriminator)
        real_name = getattr(user, "nick", None) or user.name
        # Avatar URL is an Asset object, URL only available via __str__.
        avatar = str(user.avatar_url) if user.avatar_url else None
        return cls(id_=user.id,
                   plug=discord,
                   username=username,
                   real_name=real_name,
                   avatar=avatar,
                   raw=user)


class DiscordRichText(immp.RichText):

    tags = {"**": "bold", "*": "italic", "_": "italic", "__": "underline", "~~": "strike",
            "`": "code", "```": "pre"}
    # A rather complicated expression to match formatting tags according to the following rules:
    # 1) Outside of formatting may not be adjacent to alphanumeric or other formatting characters.
    # 2) Inside of formatting may not be adjacent to whitespace or the current formatting character.
    # 3) Formatting characters may be escaped with a backslash.
    # This still isn't perfect, but provides a good approximation outside of edge cases.
    _outside_chars = r"0-9a-z*_~"
    _tag_chars = r"```|\*\*|__|~~|[*_`]"
    _inside_chars = r"\s\1"
    _format_regex = re.compile(r"(?<![{0}\\])({1})(?![{2}])(.+?)(?<![{2}\\])\1(?![{0}])"
                               .format(_outside_chars, _tag_chars, _inside_chars))

    _mention_regex = re.compile(r"<@!?(\d+)>")
    _channel_regex = re.compile(r"<#(\d+)>")
    _emoji_regex = re.compile(r"<(:[^: ]+?:)\d+>")

    @classmethod
    def _sub_channel(cls, discord, match):
        return "#{}".format(discord._client.get_channel(match.group(1)).name)

    @classmethod
    def from_markdown(cls, discord, text):
        """
        Convert a string of Markdown into a :class:`.RichText`.

        Args:
            discord (.DiscordPlug):
                Related plug instance that provides the text.
            text (str):
                Markdown formatted text.

        Returns:
            .DiscordRichText:
                Parsed rich text container.
        """
        changes = defaultdict(dict)
        while True:
            match = cls._format_regex.search(text)
            if not match:
                break
            start = match.start()
            end = match.end()
            tag = match.group(1)
            # Strip the tag characters from the message.
            text = text[:start] + match.group(2) + text[end:]
            end -= 2 * len(tag)
            # Record the range where the format is applied.
            field = cls.tags[tag]
            changes[start][field] = True
            changes[end][field] = False
        for match in cls._mention_regex.finditer(text):
            user = discord._client.get_user(int(match.group(1)))
            if user:
                changes[match.start()]["mention"] = DiscordUser.from_user(discord, user)
                changes[match.end()]["mention"] = None
        segments = []
        points = list(sorted(changes.keys()))
        formatting = {}
        # Iterate through text in change start/end pairs.
        for start, end in zip([0] + points, points + [len(text)]):
            formatting.update(changes[start])
            if start == end:
                # Zero-length segment at the start or end, ignore it.
                continue
            if formatting.get("mention"):
                user = formatting["mention"]
                part = "@{}".format(user.username or user.real_name)
            else:
                part = emojize(text[start:end], use_aliases=True)
                # Strip Discord channel/emoji tags, replace with a plain text representation.
                part = cls._channel_regex.sub(partial(cls._sub_channel, discord), part)
                part = cls._emoji_regex.sub(r"\1", part)
            segments.append(immp.Segment(part, **formatting))
        return cls(segments)

    @classmethod
    def _sub_emoji(cls, discord, match):
        for emoji in discord._client.emojis:
            if emoji.name == match.group(1):
                return str(emoji)
        return ":{}:".format(match.group(1))

    @classmethod
    def to_markdown(cls, discord, rich, webhook=False):
        """
        Convert a :class:`.RichText` instance into a Markdown string.

        Args:
            discord (.DiscordPlug):
                Related plug instance to cross-reference users.
            rich (.DiscordRichText):
                Parsed rich text container.
            webhook (bool):
                ``True`` if being sent via a webhook, which allows use of hyperlinks.

        Returns:
            str:
                Markdown formatted text.
        """
        text = ""
        active = []
        for segment in rich.normalise():
            for tag in reversed(active):
                # Check all existing tags, and remove any that end at this segment.
                attr = cls.tags[tag]
                if not getattr(segment, attr):
                    text += tag
                    active.remove(tag)
            for tag, attr in cls.tags.items():
                # Add any new tags that start at this segment.
                if getattr(segment, attr) and tag not in active:
                    text += tag
                    active.append(tag)
            if segment.mention and isinstance(segment.mention.plug, DiscordPlug):
                text += "<@{}>".format(segment.mention.id)
            elif segment.link:
                if webhook:
                    text += "[{}]({})".format(segment.text, segment.link)
                elif segment.text == segment.link:
                    text += segment.text
                else:
                    text += "{} [{}]".format(segment.text, segment.link)
            else:
                text += segment.text
        for tag in reversed(active):
            # Close all remaining tags.
            text += tag
        return re.sub(r":([^: ]+):", partial(cls._sub_emoji, discord), text)

    @classmethod
    def chunk_split(cls, text):
        """
        Split long messages into parts of at most 2000 characters.

        Args:
            text (str):
                Markdown-formatted message to be sent.

        Returns:
            str list:
                Chunked message text.
        """
        parts = []
        current = []
        for line in text.splitlines():
            size = sum(len(part) + 1 for part in current)
            extra = len(line)
            if size + extra >= 2000:
                if current:
                    # The message is full, split here.
                    parts.append("\n".join(current))
                    current.clear()
                if extra >= 2000:
                    # The line itself is too long, split on whitespace instead.
                    *lines, line = wrap(line, 2000, expand_tabs=False, replace_whitespace=False)
                    parts.extend(lines)
            current.append(line)
        if current:
            parts.append("\n".join(current))
        return parts


class DiscordMessage(immp.Message):
    """
    Message originating from Discord.
    """

    @classmethod
    def from_message(cls, discord, message, edited=False, deleted=False):
        """
        Convert a :class:`discord.Message` into a :class:`.Message`.

        Args:
            discord (.DiscordPlug):
                Related plug instance that provides the event.
            message (discord.Message):
                Discord message object received from a channel.
            edited (bool):
                Whether this message comes from an edit event.
            deleted (bool):
                Whether this message comes from a delete event.

        Returns:
            .DiscordMessage:
                Parsed message object.
        """
        text = None
        user = DiscordUser.from_user(discord, message.author)
        attachments = []
        if message.content:
            text = DiscordRichText.from_markdown(discord, message.content)
        for attach in message.attachments:
            type_ = immp.File.Type.unknown
            if attach.filename.endswith((".jpg", ".png", ".gif")):
                type_ = immp.File.Type.image
            attachments.append(immp.File(title=attach.filename,
                                         type_=type_,
                                         source=attach.url))
        for embed in message.embeds:
            if embed.image.url and embed.image.url.rsplit(".", 1)[1] in ("jpg", "png", "gif"):
                attachments.append(immp.File(type_=immp.File.Type.image,
                                             source=embed.image.url))
        return immp.SentMessage(id_=message.id,
                                channel=immp.Channel(discord, message.channel.id),
                                # Timestamps are naive but in UTC.
                                at=message.created_at.replace(tzinfo=timezone.utc),
                                # Edited timestamp is blank for new messages, but updated in
                                # existing objects when the message is later edited.
                                revision=(message.edited_at or message.created_at).timestamp(),
                                edited=edited,
                                deleted=deleted,
                                text=text,
                                user=user,
                                attachments=attachments,
                                raw=message)

    @classmethod
    async def to_embed(cls, discord_, msg, reply=False):
        """
        Convert a :class:`.Message` to a message embed structure, suitable for embedding within an
        outgoing message.

        Args:
            discord_ (.DiscordPlug):
                Target plug instance for this attachment.
            msg (.Message):
                Original message from another plug or hook.
            reply (bool):
                Whether to show a reply icon instead of a quote icon.

        Returns.
            discord.Embed:
                Discord API `embed <https://discordapp.com/developers/docs/resources/channel>`_
                object.
        """
        icon = "\N{RIGHTWARDS ARROW WITH HOOK}" if reply else "\N{SPEECH BALLOON}"
        embed = discord.Embed()
        embed.set_footer(text=icon)
        if isinstance(msg, immp.Receipt):
            embed.timestamp = msg.at
        if msg.user:
            link = discord.Embed.Empty
            # Exclude platform-specific join protocol URLs.
            if (msg.user.link or "").startswith("http"):
                link = msg.user.link
            embed.set_author(name=(msg.user.real_name or msg.user.username),
                             url=link, icon_url=msg.user.avatar or discord.Embed.Empty)
        quote = None
        action = False
        if msg.text:
            quote = msg.text.clone()
            action = msg.action
        elif msg.attachments:
            count = len(msg.attachments)
            what = "{} attachment".format(count) if count > 1 else "this attachment"
            quote = immp.RichText([immp.Segment("sent {}".format(what))])
            action = True
        if quote:
            if action:
                for segment in quote:
                    segment.italic = True
            embed.description = DiscordRichText.to_markdown(discord_, quote)
        return embed


class DiscordClient(discord.Client):
    """
    Subclass of the underlying client to bind events.
    """

    def __init__(self, plug, **kwargs):
        super().__init__(**kwargs)
        self._plug = plug

    async def on_ready(self):
        async with self._plug._starting:
            self._plug._starting.notify_all()
        await self.on_resume()

    async def on_resume(self):
        if self._plug.config["playing"]:
            await self.change_presence(activity=discord.Game(self._plug.config["playing"]))

    async def on_message(self, message):
        log.debug("Received a new message")
        self._plug.queue(DiscordMessage.from_message(self._plug, message))

    async def on_message_edit(self, before, after):
        log.debug("Received an updated message")
        if before.content == after.content:
            # Text content hasn't changed -- maybe just a link unfurl embed added.
            return
        self._plug.queue(DiscordMessage.from_message(self._plug, after, edited=True))

    async def on_message_delete(self, message):
        log.debug("Received a deleted message")
        self._plug.queue(DiscordMessage.from_message(self._plug, message, deleted=True))


class DiscordPlug(immp.HTTPOpenable, immp.Plug):
    """
    Plug for a `Discord <https://discordapp.com>`_ server.
    """

    schema = _Schema.config

    network_name = "Discord"

    @property
    def network_id(self):
        return ("discord:{}".format(self._client.user.id)
                if self._client and self._client.user else None)

    def __init__(self, name, config, host):
        super().__init__(name, config, host)
        # Connection objects that need to be closed on disconnect.
        self._client = self._task = None
        self._starting = Condition()

    async def start(self):
        await super().start()
        log.debug("Starting client")
        self._client = DiscordClient(self)
        self._task = ensure_future(self._client.start(self.config["token"], bot=self.config["bot"]))
        async with self._starting:
            # Block until the client is ready.
            await self._starting.wait()

    async def stop(self):
        await super().stop()
        if self._client:
            log.debug("Closing client")
            await self._client.close()
            self._client = None

    async def user_from_id(self, id_):
        user = await self._client.fetch_user(id_)
        return DiscordUser.from_user(self, user) if user else None

    async def user_from_username(self, username):
        for guild in self._client.guilds:
            member = guild.get_member_named(username)
            if member:
                return DiscordUser.from_user(self, member)
        return None

    async def user_is_system(self, user):
        return user.id == str(self._client.user.id)

    async def public_channels(self):
        return [immp.Channel(self, channel.id) for channel in self._client.get_all_channels()
                if isinstance(channel, discord.TextChannel)]

    async def private_channels(self):
        return [immp.Channel(self, channel.id) for channel in self._client.private_channels]

    def _get_channel(self, channel):
        return self._client.get_channel(int(channel.source))

    async def channel_for_user(self, user):
        if not isinstance(user, DiscordUser):
            return None
        if not isinstance(user.raw, (discord.Member, discord.User)):
            return None
        dm = user.raw.dm_channel or (await user.raw.create_dm())
        return immp.Channel(self, dm.id)

    async def channel_title(self, channel):
        dc_channel = self._get_channel(channel)
        return dc_channel.name if isinstance(dc_channel, discord.TextChannel) else None

    async def channel_link(self, channel):
        dc_channel = self._get_channel(channel)
        if isinstance(dc_channel, discord.TextChannel):
            guild = dc_channel.guild.id
        elif isinstance(dc_channel, (discord.DMChannel, discord.GroupChannel)):
            guild = "@me"
        else:
            return None
        return "https://discordapp.com/channels/{}/{}".format(guild, dc_channel.id)

    async def channel_rename(self, channel, title):
        dc_channel = self._get_channel(channel)
        if isinstance(dc_channel, discord.TextChannel):
            await dc_channel.edit(name=title)

    async def channel_is_private(self, channel):
        dc_channel = self._get_channel(channel)
        return isinstance(dc_channel, discord.DMChannel)

    async def channel_members(self, channel):
        dc_channel = self._get_channel(channel)
        if isinstance(dc_channel, discord.TextChannel):
            return [DiscordUser.from_user(self, member) for member in dc_channel.members]
        elif isinstance(dc_channel, discord.GroupChannel):
            return [DiscordUser.from_user(self, member) for member in dc_channel.recipients]
        elif isinstance(dc_channel, discord.DMChannel):
            return [DiscordUser.from_user(self, dc_channel.me),
                    DiscordUser.from_user(self, dc_channel.recipient)]
        else:
            return []

    async def channel_history(self, channel, before=None):
        dc_channel = self._get_channel(channel)
        dc_before = await dc_channel.fetch_message(before.id) if before else None
        history = dc_channel.history(before=dc_before, oldest_first=False)
        messages = [DiscordMessage.from_message(self, message) async for message in history]
        return list(reversed(messages))

    def _resolve_channel(self, channel):
        dc_channel = self._get_channel(channel)
        webhook = None
        for label, host_channel in self.host.channels.items():
            if channel == host_channel and label in self.config["webhooks"]:
                adapter = discord.AsyncWebhookAdapter(self.session)
                webhook = discord.Webhook.from_url(self.config["webhooks"][label], adapter=adapter)
                break
        return dc_channel, webhook

    async def _resolve_message(self, dc_channel, msg):
        if isinstance(msg, immp.Receipt):
            # Discord offers no reply mechanism, so instead we just fetch the referenced message
            # and render it manually.
            message = await dc_channel.fetch_message(msg.id)
            return DiscordMessage.from_message(self, message)
        elif isinstance(msg, immp.Message):
            return msg

    async def _requests(self, dc_channel, webhook, msg):
        name = image = None
        embeds = []
        files = []
        if msg.user:
            name = msg.user.real_name or msg.user.username
            image = msg.user.avatar
        for i, attach in enumerate(msg.attachments or []):
            if isinstance(attach, immp.File) and attach.type == immp.File.Type.image:
                async with (await attach.get_content(self.session)) as img_content:
                    # discord.py expects a file-like object with a synchronous read() method.
                    # NB. The whole file is read into memory by discord.py anyway.
                    files.append(discord.File(BytesIO(await img_content.read()),
                                              attach.title or "image_{}.png".format(i)))
            elif isinstance(attach, immp.Location):
                embed = discord.Embed()
                embed.title = attach.name or "Location"
                embed.url = attach.google_map_url
                embed.description = attach.address
                embed.set_thumbnail(url=attach.google_image_url(80))
                embed.set_footer(text="{}, {}".format(attach.latitude, attach.longitude))
                embeds.append((embed, "sent a location"))
            elif isinstance(attach, immp.Message):
                resolved = await self._resolve_message(dc_channel, attach)
                embeds.append((await DiscordMessage.to_embed(self, resolved), "sent a message"))
        if msg.reply_to:
            resolved = await self._resolve_message(dc_channel, msg.reply_to)
            embeds.append((await DiscordMessage.to_embed(self, resolved, True), None))
        edited = msg.edited if isinstance(msg, immp.Receipt) else False
        if webhook:
            # Sending via webhook: multiple embeds and files supported.
            requests = []
            text = None
            if msg.text:
                rich = msg.text.clone()
                if msg.action:
                    for segment in rich:
                        segment.italic = True
                if edited:
                    rich.append(immp.Segment(" (edited)", italic=True))
                lines = DiscordRichText.chunk_split(DiscordRichText.to_markdown(self, rich, True))
                if len(lines) > 1:
                    # Multiple messages required to accommodate the text.
                    requests.extend(webhook.send(content=line, wait=True, username=name,
                                                 avatar_url=image) for line in lines)
                else:
                    text = lines[0]
            if text or embeds or files:
                requests.append(webhook.send(content=text, wait=True, username=name,
                                             avatar_url=image, files=files,
                                             embeds=[embed[0] for embed in embeds]))
            return requests
        else:
            # Sending via client: only a single embed per message.
            requests = []
            text = embed = None
            rich = msg.render(link_name=False, edit=edited) or None
            if rich:
                lines = DiscordRichText.chunk_split(DiscordRichText.to_markdown(self, rich))
                if len(lines) > 1:
                    # Multiple messages required to accommodate the text.
                    requests.extend(dc_channel.send(content=line) for line in lines)
                else:
                    text = lines[0]
            if len(embeds) == 1:
                # Attach the only embed to the message text.
                embed, _ = embeds.pop()
            if text or embed or files:
                requests.append(dc_channel.send(content=text, embed=embed, files=files))
            for embed, desc in embeds:
                # Send any additional embeds in their own separate messages.
                content = None
                if msg.user and desc:
                    label = immp.Message(user=msg.user, text="sent {}".format(desc), action=True)
                    content = DiscordRichText.to_markdown(self, label.render())
                requests.append(dc_channel.send(content=content, embed=embed))
            return requests

    async def put(self, channel, msg):
        dc_channel, webhook = self._resolve_channel(channel)
        requests = await self._requests(dc_channel, webhook, msg)
        sent = []
        for request in requests:
            sent.append(await request)
        return [str(resp.id) for resp in sent]

    async def delete(self, sent):
        dc_channel = self._resolve_channel(sent.channel)[0]
        if not dc_channel:
            raise DiscordAPIError("No access to channel {}".format(sent.channel.source))
        message = await dc_channel.fetch_message(sent.id)
        # If not self-posted (including webhooks), the Manage Messages permission is required.
        await message.delete()



"""
Web-based management UI for a host instance.

Config:
    route (str):
        Path to expose the UI pages under.

This is a simple control panel to add or update plugs and hooks in a running system.

.. warning::
    The UI provides no authentication, and exposes all internals of the system (including any
    secret keys, as well as remote code execution via loading new plugs/hooks).  You should only
    run this hook on a trusted system, and bind it to localhost to avoid external access.

    If remotely accessible configuration is desired, use SSH with port forwarding to securely
    connect to your server.  Failing that, you could proxy via a webserver like nginx and enable
    HTTP authentication, or use a service layer like oauth2_proxy.
"""

from asyncio import gather
from collections import defaultdict
from functools import wraps
import json
import re
import logging

from aiohttp import web
import aiohttp_jinja2

import immp
from immp.hook.runner import RunnerHook
from immp.hook.web import WebHook


log = logging.getLogger(__name__)


class WebUIHook(immp.ResourceHook):
    """
    Hook providing web-based configuration management for a running host instance.
    """

    schema = immp.Schema({"route": str})

    def __init__(self, name, config, host):
        super().__init__(name, config, host)
        self.ctx = None

    def on_load(self):
        log.debug("Registering routes")
        runner = self.host.resources.get(RunnerHook)
        self.ctx = self.host.resources[WebHook].context(self.config["route"], __name__,
                                                        env={"hook_url_for": self.hook_url_for,
                                                             "group_summary": self.group_summary,
                                                             "runner": runner,
                                                             # `zip` doesn't seem to work.
                                                             "zipped": zip})
        # Home:
        self.ctx.route("GET", "", self.noop, "main.j2", "main")
        # Add:
        self.ctx.route("GET", "add", self.noop, "add.j2", "add")
        self.ctx.route("POST", "add", self.add, "add.j2", "add:post")
        # Plugs:
        self.ctx.route("GET", "plug/{name}", self.plug, "plug.j2")
        self.ctx.route("POST", "plug/{name}/stop", self.plug_stop)
        self.ctx.route("POST", "plug/{name}/start", self.plug_start)
        self.ctx.route("POST", "plug/{name}/config", self.plug_config)
        self.ctx.route("GET", "plug/{name}/channels", self.plug_channels, "plug_channels.j2")
        self.ctx.route("GET", "plug/{name}/remove", self.plug, "plug_remove.j2", "plug_remove")
        self.ctx.route("POST", "plug/{name}/remove", self.plug_remove, name="plug_remove:post")
        # Channels:
        self.ctx.route("GET", "channel/{name}", self.named_channel, "channel.j2")
        self.ctx.route("POST", "channel", self.named_channel_add)
        self.ctx.route("POST", "channel/{name}/remove", self.named_channel_remove)
        self.ctx.route("GET", "plug/{plug}/channel/{source}", self.channel, "channel.j2")
        self.ctx.route("POST", "plug/{plug}/channel/{source}/migrate", self.channel_migration)
        self.ctx.route("POST", "plug/{plug}/channel/{source}/invite", self.channel_invite)
        self.ctx.route("POST", "plug/{plug}/channel/{source}/kick/{user}", self.channel_kick)
        # Groups:
        self.ctx.route("GET", "group/{name}", self.group, "group.j2")
        self.ctx.route("POST", "group", self.group_add)
        self.ctx.route("POST", "group/{name}/remove", self.group_remove)
        self.ctx.route("POST", "group/{name}/config", self.group_config)
        # Hooks:
        self.ctx.route("GET", "resource/{cls}", self.hook, "hook.j2", "resource")
        self.ctx.route("POST", "resource/{cls}/stop", self.hook_stop, name="resource_stop")
        self.ctx.route("POST", "resource/{cls}/start", self.hook_start, name="resource_start")
        self.ctx.route("POST", "resource/{cls}/config", self.hook_config, name="resource_config")
        self.ctx.route("GET", "resource/{cls}/remove", self.hook,
                       "hook_remove.j2", "resource_remove")
        self.ctx.route("POST", "resource/{cls}/remove", self.hook_remove,
                       name="resource_remove:post")
        self.ctx.route("GET", "hook/{name}", self.hook, "hook.j2")
        self.ctx.route("POST", "hook/{name}/stop", self.hook_stop)
        self.ctx.route("POST", "hook/{name}/start", self.hook_start)
        self.ctx.route("POST", "hook/{name}/config", self.hook_config)
        self.ctx.route("GET", "hook/{name}/remove", self.hook, "hook_remove.j2", "hook_remove")
        self.ctx.route("POST", "hook/{name}/remove", self.hook_remove, name="hook_remove:post")

    async def noop(self, request):
        return {}

    async def add(self, request):
        post = await request.post()
        try:
            path = post["path"]
        except KeyError:
            raise web.HTTPBadRequest
        if not path:
            raise web.HTTPBadRequest
        try:
            cls = immp.resolve_import(path)
        except ImportError:
            raise web.HTTPNotFound
        if "schema" in post:
            config = post.get("config") or ""
            return {"path": path,
                    "config": config,
                    "class": cls,
                    "hook": issubclass(cls, immp.Hook)}
        try:
            name = post["name"]
        except KeyError:
            raise web.HTTPBadRequest
        if not name:
            raise web.HTTPBadRequest
        elif name in self.host:
            raise web.HTTPConflict
        if cls.schema:
            try:
                config = json.loads(post["config"])
            except (KeyError, ValueError):
                raise web.HTTPBadRequest
        else:
            config = None
        if not issubclass(cls, (immp.Plug, immp.Hook)):
            raise web.HTTPNotFound
        try:
            inst = cls(name, config, self.host)
        except immp.Invalid:
            raise web.HTTPNotAcceptable
        if issubclass(cls, immp.Plug):
            self.host.add_plug(inst)
            raise web.HTTPFound(self.ctx.url_for("plug", name=name))
        elif issubclass(cls, immp.Hook):
            try:
                priority = int(post["priority"]) if post["priority"] else None
            except (KeyError, ValueError):
                raise web.HTTPBadRequest
            self.host.add_hook(inst, priority)
            if issubclass(cls, immp.ResourceHook):
                raise web.HTTPFound(self.ctx.url_for("resource", cls=path))
            else:
                raise web.HTTPFound(self.ctx.url_for("hook", name=name))

    def _resolve_plug(self, request):
        try:
            return self.host.plugs[request.match_info["name"]]
        except KeyError:
            raise web.HTTPNotFound

    async def plug(self, request):
        plug = self._resolve_plug(request)
        name = None
        source = request.query.get("source")
        if source:
            title = await immp.Channel(plug, source).title()
            name = re.sub(r"[^a-z0-9]+", "-", title, flags=re.I).strip("-") if title else ""
        return {"plug": plug,
                "add_name": name,
                "add_source": source,
                "channels": {name: channel for name, channel in self.host.channels.items()
                             if channel.plug == plug}}

    async def plug_stop(self, request):
        plug = self._resolve_plug(request)
        await plug.close()
        raise web.HTTPFound(self.ctx.url_for("plug", name=plug.name))

    async def plug_start(self, request):
        plug = self._resolve_plug(request)
        await plug.open()
        raise web.HTTPFound(self.ctx.url_for("plug", name=plug.name))

    async def plug_config(self, request):
        plug = self._resolve_plug(request)
        if not plug.schema:
            raise web.HTTPNotFound
        post = await request.post()
        if "config" not in post:
            raise web.HTTPBadRequest
        try:
            config = json.loads(post["config"])
        except ValueError:
            raise web.HTTPNotAcceptable
        try:
            plug.schema(config)
        except immp.Invalid:
            raise web.HTTPNotAcceptable
        plug.config.clear()
        plug.config.update(config)
        raise web.HTTPFound(self.ctx.url_for("plug", name=plug.name))

    async def plug_channels(self, request):
        plug = self._resolve_plug(request)
        public, private = await gather(plug.public_channels(), plug.private_channels())
        titles = await gather(*(channel.title() for channel in public)) if public else []
        all_members = await gather(*(channel.members() for channel in private)) if private else []
        users = []
        for members in all_members:
            if not members:
                users.append([])
                continue
            systems = await gather(*(member.is_system() for member in members)) if members else []
            users.append([member for member, system in zip(members, systems) if not system])
        channels = defaultdict(list)
        for name, channel in self.host.channels.items():
            if channel.plug == plug:
                channels[channel].append(name)
        return {"plug": plug,
                "channels": channels,
                "public": public,
                "titles": titles,
                "private": private,
                "users": users}

    async def plug_remove(self, request):
        plug = self._resolve_plug(request)
        await plug.stop()
        self.host.remove_plug(plug.name)
        raise web.HTTPFound(self.ctx.url_for("main"))

    def _resolve_channel(self, request):
        try:
            if "name" in request.match_info:
                name = request.match_info["name"]
                return name, self.host.channels[name]
            elif "plug" in request.match_info:
                plug = self.host.plugs[request.match_info["plug"]]
                return None, immp.Channel(plug, request.match_info["source"])
            else:
                raise web.HTTPBadRequest
        except KeyError:
            raise web.HTTPNotFound

    async def named_channel(self, request):
        name, channel = self._resolve_channel(request)
        raise web.HTTPFound(self.ctx.url_for("channel", plug=channel.plug.name,
                                             source=channel.source))

    async def named_channel_add(self, request):
        post = await request.post()
        try:
            plug = post["plug"]
            name = post["name"]
            source = post["source"]
        except KeyError:
            raise web.HTTPBadRequest
        if not (plug and name and source):
            raise web.HTTPBadRequest
        if name in self.host:
            raise web.HTTPConflict
        if plug not in self.host.plugs:
            raise web.HTTPNotFound
        self.host.add_channel(name, immp.Channel(self.host.plugs[plug], source))
        raise web.HTTPFound(self.ctx.url_for("plug", name=plug))

    async def named_channel_remove(self, request):
        name, channel = self._resolve_channel(request)
        self.host.remove_channel(name)
        raise web.HTTPFound(self.ctx.url_for("plug", name=channel.plug.name))

    async def channel(self, request):
        name, channel = self._resolve_channel(request)
        private = await channel.is_private()
        title = await channel.title()
        link = await channel.link()
        members = await channel.members()
        return {"name": name,
                "channel": channel,
                "private": private,
                "title_": title,
                "link": link,
                "members": members}

    async def channel_migration(self, request):
        name, old = self._resolve_channel(request)
        post = await request.post()
        if "name" in post:
            new = self.host.channels[post["name"]]
        elif "plug" in post and "source" in post:
            new = immp.Channel(self.host.plugs[post["plug"]], post["source"])
        else:
            raise web.HTTPBadRequest
        hooks = await self.host.channel_migrate(old, new)
        summary = ",".join(hooks)
        raise web.HTTPFound(self.ctx.url_for("channel", plug=old.plug.name, source=old.source))

    async def channel_invite(self, request):
        name, channel = self._resolve_channel(request)
        if channel.plug.virtual:
            raise web.HTTPBadRequest
        post = await request.post()
        try:
            id_ = post["user"]
        except KeyError:
            raise web.HTTPBadRequest
        members = await channel.members()
        if members is None:
            raise web.HTTPBadRequest
        elif id_ in (member.id for member in members):
            raise web.HTTPBadRequest
        user = await channel.plug.user_from_id(id_)
        if user is None:
            raise web.HTTPBadRequest
        await channel.invite(user)
        raise web.HTTPFound(self.ctx.url_for("channel", plug=channel.plug.name,
                                             source=channel.source))

    async def channel_kick(self, request):
        name, channel = self._resolve_channel(request)
        if channel.plug.virtual:
            raise web.HTTPBadRequest
        id_ = request.match_info["user"]
        members = await channel.members()
        if members is None:
            raise web.HTTPBadRequest
        elif id_ not in (member.id for member in members):
            raise web.HTTPBadRequest
        user = await channel.plug.user_from_id(id_)
        if user is None:
            raise web.HTTPBadRequest
        await channel.remove(user)
        raise web.HTTPFound(self.ctx.url_for("channel", plug=channel.plug.name,
                                             source=channel.source))

    def _resolve_group(self, request):
        try:
            return self.host.groups[request.match_info["name"]]
        except KeyError:
            raise web.HTTPNotFound

    @staticmethod
    def group_summary(group):
        summary = []
        for key in ("anywhere", "private", "shared", "named", "channels"):
            count = len(group.config[key])
            if count:
                summary.append("{} {}".format(count, key))
        return ", ".join(summary) if summary else "Empty group"

    async def group(self, request):
        group = self._resolve_group(request)
        return {"group": group}

    async def group_add(self, request):
        post = await request.post()
        try:
            name = post["name"]
        except KeyError:
            raise web.HTTPBadRequest
        if not name:
            raise web.HTTPBadRequest
        if name in self.host:
            raise web.HTTPConflict
        self.host.add_group(immp.Group(name, {}, self.host))
        raise web.HTTPFound(self.ctx.url_for("group", name=name))

    async def group_remove(self, request):
        group = self._resolve_group(request)
        self.host.remove_group(group.name)
        raise web.HTTPFound(self.ctx.url_for("main"))

    async def group_config(self, request):
        group = self._resolve_group(request)
        post = await request.post()
        if "config" not in post:
            raise web.HTTPBadRequest
        try:
            config = json.loads(post["config"])
        except ValueError:
            raise web.HTTPBadRequest
        group.config.clear()
        group.config.update(config)
        raise web.HTTPFound(self.ctx.url_for("group", name=group.name))

    def _resolve_hook(self, request):
        if "name" in request.match_info:
            try:
                return self.host.hooks[request.match_info["name"]]
            except KeyError:
                pass
        elif "cls" in request.match_info:
            for cls, hook in self.host.resources.items():
                if request.match_info["cls"] == "{}.{}".format(cls.__module__, cls.__name__):
                    return hook
        raise web.HTTPNotFound

    def hook_url_for(self, hook, name_=None, **kwargs):
        if isinstance(hook, immp.ResourceHook):
            cls = "{}.{}".format(hook.__class__.__module__, hook.__class__.__name__)
            route = "resource_{}".format(name_) if name_ else "resource"
            return self.ctx.url_for(route, cls=cls, **kwargs)
        else:
            route = "hook_{}".format(name_) if name_ else "hook"
            return self.ctx.url_for(route, name=hook.name, **kwargs)

    async def hook(self, request):
        hook = self._resolve_hook(request)
        can_stop = not isinstance(hook, (WebHook, WebUIHook))
        return {"hook": hook,
                "resource": isinstance(hook, immp.ResourceHook),
                "priority": self.host.priority.get(hook.name),
                "can_stop": can_stop}

    async def hook_stop(self, request):
        hook = self._resolve_hook(request)
        if isinstance(hook, (WebHook, WebUIHook)):
            # This will hang due to trying to serve this request at the same time.
            raise web.HTTPBadRequest
        await hook.close()
        raise web.HTTPFound(self.hook_url_for(hook, None))

    async def hook_start(self, request):
        hook = self._resolve_hook(request)
        await hook.open()
        raise web.HTTPFound(self.hook_url_for(hook, None))

    async def hook_config(self, request):
        hook = self._resolve_hook(request)
        post = await request.post()
        try:
            priority = int(post["priority"]) if post["priority"] else None
        except (KeyError, ValueError):
            raise web.HTTPBadRequest
        self.host.prioritise_hook(hook.name, priority)
        if hook.schema:
            if "config" not in post:
                raise web.HTTPBadRequest
            try:
                config = json.loads(post["config"])
            except ValueError:
                raise web.HTTPNotAcceptable
            try:
                hook.schema(config)
            except immp.Invalid:
                raise web.HTTPNotAcceptable
            hook.config.clear()
            hook.config.update(config)
        raise web.HTTPFound(self.hook_url_for(hook, None))

    async def hook_remove(self, request):
        hook = self._resolve_hook(request)
        await hook.stop()
        self.host.remove_hook(hook.name)
        raise web.HTTPFound(self.ctx.url_for("main"))
